/* 
 * am.c
 *
 * Implementation of most of the AM interface functions.
 */
#include "am.h"
#include "stdam.h"
#include "cMutex.h"
#include "cAMLayer.h"
#include "cEndpoint.h"
#include "cBundle.h"
#include "cConnection.h"
#include "Header.h"

//#define DEBUG_INFO

/* Global Vars */
cAMLayer	*gAMLayer;
ostream		*gAMErrorStream = 0;
cMutex		 gAMMutex;

/* Func Prototypes */
void AM_SetErrorStream(ostream &errStream);

/*
 * AM_Init()
 *
 * Purpose:	Create and initialize the active message layer.
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer is not initialized.
 * PostCnd:	The AM Layer is created and initialized.
 * Return:	AM_OK if ok, else AM_ERR_NOT_INIT
 */
int	AM_Init()
{
	// <LOCK>
	gAMMutex.Lock();

	// Set the VI error stream.
	cVIError::SetErrorStream(cerr);
	
	if(gAMLayer)
	{
		gAMMutex.Unlock();
		return AM_OK;
	}

	gAMLayer = new cAMLayer();

	if (!gAMLayer)
	{
		AM_Error("Cannot make cAMLayer object.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		gAMMutex.Unlock();
		return AM_ERR_NOT_INIT;
	}

	if(!gAMLayer->Initialize("nic0"))
	{
		AM_Error("Cannot initialize gAMLayer object.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		delete gAMLayer;
		gAMLayer = NULL;
		gAMMutex.Unlock();
		return AM_ERR_NOT_INIT;
	}

	// <UNLOCK>
	gAMMutex.Unlock();

	return AM_OK;
}

/*
 * AM_Terminate()
 *
 * Purpose:	Shut down the AM Layer.
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer was previously initialized.
 * PostCnd:	The AM Layer is destroyed and null.
 * Return:	AM_OK if ok, else AM_ERR_NOT_INIT
 */
int	AM_Terminate()
{
	// <LOCK>
	gAMMutex.Lock();

	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		gAMMutex.Unlock();
		return AM_ERR_NOT_INIT;
	}

	delete gAMLayer;
	gAMLayer = NULL;

	// <UNLOCK>
	gAMMutex.Unlock();
	return AM_OK;
}

/*
 * AM_Request4()
 *
 * Purpose:	Invoke a handler on a remote endpoint (w/ 4 args).
 * IN:		request_endpoint -> endpoint sending request.
 *			reply_endpoint	 -> endpoint sending reply (ep to send to).
 *			handler			 -> index into dest handler table.
 *			a0-a3			 -> integer args to remote handler.
 * OUT:		-
 * Cond:	Request handler is void handler(void *token, int a0, ... int a3).
 * PostCnd:	The handler on the remote endpoint should execute.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int AM_Request4(ep_t request_endpoint,			
				int reply_endpoint,
				handler_t handler,				
				int a0, int a1, int a2, int a3	
				)
{
	// Sanity check
	if(!request_endpoint)
	{
		AM_Error("Request4(): Bad endpoint arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}
	return request_endpoint->Request4(reply_endpoint, handler, a0, a1, a2, a3);
}

/*
 * AM_RequestI4()
 *
 * Purpose:	Invoke a handler on a remote endpoint (w/ 4 args and buffer).
 * IN:		request_endpoint -> endpoint sending request.
 *			reply_endpoint	 -> endpoint sending reply (ep to send to).
 *			handler			 -> index into dest handler table.
 *			source_addr		 -> Address of source buffer.
 *			nbytes			 -> Lenght of source buffer.
 *			a0-a3			 -> integer args to remote handler.
 * OUT:		-
 * Cond:	Request handler is void handler(void *token, void *buf, int nbytes, int a0, ... int a3).
 * PostCnd:	The handler on the remote endpoint should execute.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int AM_RequestI4(ep_t request_endpoint,			
				int reply_endpoint,				
				handler_t handler,				
				void *source_addr,				
				int nbytes,						
				int a0, int a1, int a2, int a3	
				)
{
	// Sanity check
	if(!request_endpoint)
	{
		AM_Error("RequestI4(): Bad endpoint arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!source_addr)
	{
		AM_Error("RequestI4(): Bad buffer source address.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}
	return request_endpoint->RequestI4(reply_endpoint, handler, source_addr, nbytes, a0, a1, a2, a3);
}

/*
 * AM_RequestXfer4()
 *
 * Purpose:	Invoke a handler on a remote endpoint (w/ 4 args and data copied into virtual mem segment).
 * IN:		request_endpoint -> endpoint sending request.
 *			reply_endpoint	 -> endpoint sending reply (ep to send to).
 *			dest_offset		 -> offset into remote ep's virtual memory segment.
 *			handler			 -> index into dest handler table.
 *			source_addr		 -> Address of source buffer.
 *			nbytes			 -> Lenght of source buffer.
 *			a0-a3			 -> integer args to remote handler.
 * OUT:		-
 * Cond:	Request handler is void handler(void *token, void *buf, int nbytes, int a0, ... int a3).
 * PostCnd:	The handler on the remote endpoint should execute.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int AM_RequestXfer4(ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void *source_addr, int nbytes, int a0, int a1, int a2, int a3)
{
	// Sanity check
	if(!request_endpoint)
	{
		AM_Error("RequestI4(): Bad endpoint arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!source_addr)
	{
		AM_Error("RequestI4(): Bad buffer source address.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}
	return request_endpoint->RequestXfer4(reply_endpoint, dest_offset, handler, source_addr, nbytes, a0, a1, a2, a3);
}


/*
 * AM_Reply4()
 *
 * Purpose:	Reply from a handler (w/ 4 args).
 * IN:		token			 -> the same token that was given to handler.
 *			handler			 -> index into dest handler table.
 *			a0-a3			 -> integer args to remote handler.
 * OUT:		-
 * Cond:	Request handler is void handler(void *token, int a0, ... int a3).
 * PostCnd:	The handler on the remote endpoint should execute.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int	AM_Reply4(void *token,						
			  handler_t handler,				
			  int a0, int a1, int a2, int a3	
			  )
{
	// Sanity check
	if(!token)
	{
		AM_Error("Reply4(): Bad token arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}

	cEndpoint *temp;
	if(AM_GetDestEndpoint(token, &temp) == AM_OK)
	{
		return temp->Reply4(token, handler, a0, a1, a2, a3);
	}
	else
	{
		return AM_ERR_BAD_ARG;	// Bad token arg.
	}
}

/*
 * AM_ReplyI4()
 *
 * Purpose:	Reply from a handler (w/ 4 args and data buffer).
 * IN:		token			 -> the same token that was given to handler.
 *			handler			 -> index into dest handler table.
 *			source_addr		 -> buffer to send in reply.
 *			nbytes			 -> number of bytes to send from buffer.
 *			a0-a3			 -> integer args to remote handler.
 * OUT:		-
 * Cond:	Request handler is void handler(void *token, int a0, ... int a3).
 * PostCnd:	The handler on the remote endpoint should execute.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int	AM_ReplyI4(void *token, handler_t handler, void *source_addr, int nbytes, int a0, int a1, int a2, int a3)
{
	// Sanity check
	if(!token)
	{
		AM_Error("Reply4(): Bad token arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}
	if(!source_addr)
	{
		AM_Error("ReplyI4(): Bad source_addr arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}

	cEndpoint *temp;
	if(AM_GetDestEndpoint(token, &temp) == AM_OK)
	{
		return temp->ReplyI4(token, handler, source_addr, nbytes, a0, a1, a2, a3);
	}
	else
	{
		return AM_ERR_BAD_ARG;	// Bad token arg.
	}
}

/*
 * AM_ReplyXfer4()
 *
 * Purpose:	Reply from a handler (w/ 4 args and data buffer into remote vm segment).
 * IN:		token			 -> the same token that was given to handler.
 *			dest_offset		 -> offset into remote segment to place data.
 *			handler			 -> index into dest handler table.
 *			source_addr		 -> buffer to send in reply.
 *			nbytes			 -> number of bytes to send from buffer.
 *			a0-a3			 -> integer args to remote handler.
 * OUT:		-
 * Cond:	Reply handler is void handler(void *token, void *buf, int nbytes, void *int a0, ... int a3).
 * PostCnd:	The handler on the remote endpoint should execute.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int AM_ReplyXfer4(void *token, int dest_offset, handler_t handler, void *source_addr, int nbytes, int a0, int a1, int a2, int a3)
{
	// Sanity check
	if(!token)
	{
		AM_Error("ReplyXfer4(): Bad token arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM not initialized.", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}
	if(!source_addr)
	{
		AM_Error("ReplyXfer4(): Bad source_addr arg.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}

	cEndpoint *temp;
	if(AM_GetDestEndpoint(token, &temp) == AM_OK)
	{
		return temp->ReplyXfer4(token, dest_offset, handler, source_addr, nbytes, a0, a1, a2, a3);
	}
	else
	{
		return AM_ERR_BAD_ARG;	// Bad token arg.
	}
}

/*
 * AM_GetSourceEndpoint()
 *
 * Purpose:	Find the remote endpoint name that was the source of a request.
 * IN:		token	-> The token that was given to the handler.
 * OUT:		gan		-> The address of the source.
 * Cond:	token is a valid token given to a handler.
 * PostCnd:	gan is alloced and should later be "delete"d.
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int AM_GetSourceEndpoint(void *token, en_t *gan)
{
	// Sanity
	if(!token || !gan)
	{
		return AM_ERR_BAD_ARG;
	}

	Token *ptoken;
	ptoken = (Token *)token;

	(ptoken->connection->GetAMAddress())->endpoint = ptoken->source;
	*gan = ptoken->connection->GetAMAddress();
	return AM_OK;
}

/*
 * AM_GetDestEndpoint()
 *
 * Purpose:	Find the local endpoint that is the dest of a request.
 * IN:		token	-> The token that was given to the handler.
 * OUT:		enp		-> The address of the local endpoint.
 * Cond:	token is a valid token given to a handler.
 * PostCnd:	-
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int	AM_GetDestEndpoint(void *token, ep_t	*endp)
{
	// Sanity
	if(!token || !endp)
	{
		return AM_ERR_BAD_ARG;
	}

	Token *ptoken;
	ptoken = (Token *)token;

	*endp = ptoken->endp;
	return AM_OK;
}
/*
 * AM_GetMsgTag()
 *
 * Purpose:	Retrieves the tag from the given token.
 * IN:		token	-> The token that was given to the handler.
 * OUT:		tagp	-> Filled with the tag from the token.
 * Cond:	token is a valid token given to a handler.
 * PostCnd:	-
 * Return:	AM_OK if success, else { AM_ERR_* }
 */
int AM_GetMsgTag(void *token, tag_t *tagp)
{
	// Sanity
	if(!token || !tagp)
	{
		return AM_ERR_BAD_ARG;
	}

	Token *ptoken;
	ptoken = (Token *)token;

	*tagp = ptoken->tag;
	return AM_OK;
}

/*
 * AM_AllocateBundle()
 *
 * Purpose:	Allocate an endpoint bundle.
 * IN:		type			-> AM_SEQ | AM_PAR
 * OUT:		endb			-> Contains address of the new bundle (if success).
 * Cond:	AM layer init.
 * PostCnd:	The bundle is allocated and local endpoint name is returned.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int	AM_AllocateBundle(int type, eb_t *endb)
{
	cBundle			*newBundle;

	// Sanity
	if((type != AM_SEQ) && (type != AM_PAR))
	{
		AM_Error("AM_AllocateBundle().", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	if(!gAMLayer)
	{
		AM_Error("AM_AllocateBundle().", __LINE__, __FILE__, AM_ERR_NOT_INIT);
		return AM_ERR_NOT_INIT;
	}

	// Alloc the new bundle.
	newBundle = gAMLayer->AllocateBundle();
	if(!newBundle)
	{
		*endb = NULL;
		AM_Error("AM_AllocateBundle(): Unable to allocate a bundle.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		return AM_ERR_RESOURCE;
	}
	*endb = newBundle;

	return AM_OK;
}

/*
 * AM_AllocateEndpoint()
 *
 * Purpose:	Allocate an endpoint in the given bundle.
 * IN:		bundle	-> Bundle to create endpoint in.
 * OUT:		enp		-> Contains address of new endpoint.
 *			endpoint_name	-> Points to unique endpoint name.
 * Cond:	bundle is a valid bundle.
 * PostCnd:	A new endpoint is allocated in the given bundle.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int	AM_AllocateEndpoint(eb_t bundle, ep_t *enp, en_t *endpoint_name)
{
	int				 retVal;

	// Sanity checks:
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!bundle)
	{
		return AM_ERR_BAD_ARG;
	}

	// Allocate the endpoint.
	retVal = bundle->AllocateEndpoint(enp, endpoint_name);
	if (retVal != AM_OK)
	{
		*enp = NULL;
		*endpoint_name = NULL;
		return retVal;
	}
	return  AM_OK;
}

/*
 * AM_SetExpectedResources()
 *
 * Purpose:	Requests the system alloc resources for the expected number of connections
 *			and outstanding requests per connection.
 * IN:		ea						-> Endpoint to set resources on.
 *			n_endpoints				-> Number of endpoints that this ep should be able to connect to.
 *			n_outstanding_requests	-> Number of outstanding requests per endpoint.
 * OUT:		-
 * Cond:	Ideally this should be called before any connections are made.
 * PostCnd:	The resources are 'reserved' if successful.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int AM_SetExpectedResources(ep_t ea, int n_endpoints, int n_outstanding_requests)
{
	// Sanity check.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->SetExpectedResources(n_endpoints, n_outstanding_requests);
}

/*
 * AM_Map()
 *
 * Purpose:	Map a remote address into the endpoints map table.
 * IN:		ea		-> Endpoint to change address table in.
 *			index	-> Index in map table to change.
 *			endpoint-> Name of remote endpoint to map to.
 *			tag		-> Tag to be used when sending to remote endpoint.
 * OUT:		-
 * Cond:	ea is a valid endpoint.
 * PostCnd:	ea's address map table is updated/changed.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int	AM_Map(ep_t ea, int index, en_t endpoint, tag_t tag)
{
	// Sanity check.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!ea || !endpoint)
	{
		return AM_ERR_BAD_ARG;
	}

	return ea->Map(index, endpoint, tag);
}

/*
 * AM_MapAny()
 *
 * Purpose:	Map a remote address into the endpoints map table.
 * IN:		ea		-> Endpoint to change address table in.
 *			endpoint-> Name of remote endpoint to map to.
 *			tag		-> Tag to be used when sending to remote endpoint.
 * OUT:		index	-> Index to which the remote endpoint was mapped.
 * Cond:	ea is a valid endpoint.
 * PostCnd:	ea's address map table is updated/changed.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int AM_MapAny(ep_t ea, int *index, en_t endpoint, tag_t tag)
{
	// Sanity check.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!ea || !endpoint)
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->MapAny(index, endpoint, tag);
}

/*
 * AM_UnMap()
 *
 * Purpose:	UnMap a remote address in the endpoints map table.
 * IN:		ea		-> Endpoint to change address table in.
 *			index	-> Index in map table to change.
 * OUT:		-
 * Cond:	ea is a valid endpoint.
 * PostCnd:	ea's address map table is cleared at index
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int AM_Unmap(ep_t ea, int index)
{
	//Sanity checks.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!ea)
	{
		return AM_ERR_BAD_ARG;
	}

	return ea->Unmap(index);
}

/*
 * AM_FreeEndpoint()
 *
 * Purpose:	Deletes the given endpoint and frees resources.
 * IN:		ea		-> Endpoint to free.
 * OUT:		-
 * Cond:	ea is a valid endpoint.
 * PostCnd:	ea is deleted and all resources associated with it are freed.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int	AM_FreeEndpoint(ep_t ea)
{
	cBundle	*bundle;

	// Sanity check
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!ea)
	{
		return AM_ERR_BAD_ARG;
	}

	bundle = ea->GetBundle();
	return bundle->FreeEndpoint(ea);
}


/*
 * AM_FreeBundle()
 *
 * Purpose:	Deletes the given bundle and frees resources.
 * IN:		bundle	-> Bundle to free.
 * OUT:		-
 * Cond:	bundle is a valid bundle.
 * PostCnd:	bundle is deleted and all resources associated with it are freed.
 *			NOTE: If there are any endpoints in bundle, they are freed as well!
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int	AM_FreeBundle(eb_t bundle)
{
	// Sanity.	
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if (!bundle)
	{
		return AM_ERR_BAD_ARG;
	}

	// Attempt to remove from AMLayer.
	gAMLayer->FreeBundle(bundle);

	return AM_OK;
}


/*
 * AM_SetHandler()
 *
 * Purpose:	Adds the handler to the handler table of the given endpoint.
 * IN:		ea			-> The endpoint to set the handler for.
 *			handler		-> The index into ea's handler table.
 *			function	-> Pointer to the function to be used as the handler.
 * OUT:		-
 * Cond:	ea is a valid endpoint, funciton has the appropriate form.
 * PostCnd:	index "handler" is mapped to the handler function.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int	AM_SetHandler(ep_t ea, handler_t handler, void (*function)())
{
	// Sanity check.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->SetHandler(handler, function);
}

/*
 * AM_SetHandlerAny()
 *
 * Purpose:	Adds the handler to the handler table of the given endpoint.
 * IN:		ea			-> The endpoint to set the handler for.
 *			function	-> Pointer to the function to be used as the handler.
 * OUT:		handler		-> The index into ea's handler table that this was set to.
 * Cond:	ea is a valid endpoint, functioon has the appropriate form.
 * PostCnd:	index "handler" is mapped to the handler function.
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int AM_SetHandlerAny(ep_t ea, handler_t *handler, void (*function)())
{
	// Sanity check.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->SetHandlerAny(handler, function);
}

/*
 * AM_GetNumHandlers()
 * 
 * Purpose:	Returns the size of the handler table for the given endpoint.
 * IN:		ea			-> The endpoint.
 * OUT:		n_handlers	-> Filled with the handler table size.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetNumHandlers(ep_t ea, int *n_handlers)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	*n_handlers = ea->GetNumHandlers();
	return AM_OK;
}

/*
 * AM_SetNumHandlers()
 *
 * Purpose:	Sets the handler table size if possible.
 * IN:		ea			-> The enpdoint on which to change the table size.
 *			n_handlers	-> The new number of handlers required.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The HandlerTable is modified.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int AM_SetNumHandlers(ep_t ea, int n_handlers)
{
	// Sanity
	if((!ea) || (n_handlers == 0) || (n_handlers > AM_MAX_NUM_HANDLERS))
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->SetNumHandlers((unsigned int)n_handlers);
}

/*
 * AM_Poll()	-- DEPRACATED
 * 
 * Purpose:	Services active messages from all endpoints in the bundle.
 * IN:		-
 * OUT:		-
 * Cond:	AM is initialized.
 * PostCnd:	-
 * Return:	AM_OK if success, else  { AM_ERR_* }
 */
int AM_Poll(eb_t bundle)
{
	// Sanity check.
	if (!gAMLayer)
	{
		return AM_ERR_NOT_INIT;
	}
	if(!bundle)
	{
		return AM_ERR_BAD_ARG;
	}
	return AM_OK;
}

/*
 * AM_MaxShort()
 * 
 * Purpose:	Returns maximum number of integer handler args.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	max number of int args for a handler.
 */
int AM_MaxShort()
{
	return AM_MAX_SHORT;
}

/*
 * AM_MaxMedium()
 * 
 * Purpose:	Returns MTS for a medium active message (in bytes).
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	MTS of a medium AM transmission.
 */
int AM_MaxMedium()
{
	return AM_MAX_MEDIUM;
}

/*
 * AM_MaxLong()
 * 
 * Purpose:	Returns MTS for an active xfer and get op in bytes.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	MTS for an active transfer.
 */
int AM_MaxLong()
{
	return AM_MAX_LONG;
}

/*
 * AM_SetSeg()
 * 
 * Purpose:	Sets the endpoint's virtual memory segment.
 * IN:		ea		-> endpoint on which to set the segment.
 *			nbytes	-> the size of the segment.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The given segments is used as the endpoints vm segment.
 * Return:	AM_OK or AM_ERR_*
 */
int AM_SetSeg(ep_t ea, void *addr, int nbytes)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	ea->SetSegment(addr, nbytes);
	return AM_OK;
}

/*
 * AM_GetSeg()
 * 
 * Purpose:	Gets the endpoint's virtual memory segment.
 * IN:		ea		-> endpoint on which to get the segment.
 * OUT:		nbytes	-> the size of the segment.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetSeg(ep_t ea, void **addr, int *nbytes)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	*addr   = ea->GetSegment();
	*nbytes = ea->GetSegmentSize();
	return AM_OK;
}

/*
 * AM_MaxSegLength()
 * 
 * Purpose:	Returns the maximum length a segment can be.
 * IN:		-
 * OUT:		nbytes	-> the maximum size of a segment.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_MaxSegLength(int *nbytes)
{
	*nbytes = AM_MAX_SEGMENT_SIZE;
	return AM_OK;
}

/*
 * AM_SetTag()
 * 
 * Purpose:	Sets the endpoint's tag.
 * IN:		ea		-> endpoint on which to set the tag.
 *			tag		-> the tag to set to.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_SetTag(ep_t ea, tag_t tag)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->SetTag(tag);
}

/*
 * AM_GetTag()
 * 
 * Purpose:	Gets the endpoint's tag.
 * IN:		ea		-> endpoint on which to set the tag.
 * OUT:		tag		-> the endpoint's tag.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetTag(ep_t ea, tag_t *tag)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	*tag = ea->GetTag();
	return AM_OK;
}


/*
 * AM_MaxNumTranslations()
 * 
 * Purpose:	Gets the maximum translation table index.
 * IN:		-
 * OUT:		ntrans		-> Max translation table index.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_MaxNumTranslations(int *ntrans)
{
	*ntrans = AM_MAX_NUM_TRANSLATIONS;
	return AM_OK;
}

/*
 * AM_MaxNumHandlers()
 * 
 * Purpose:	Gets the maximum number of handlers per endpoint.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_MaxNumHandlers()
{
	return AM_MAX_NUM_HANDLERS;
}

/*
 * AM_GetTranslationName()
 * 
 * Purpose:	Fills in the name of the endpoint mapped.
 * IN:		ea		-> The endpoint to search translation table of.
 *			i		-> The endpoint index number.
 * OUT:		gan		-> The remote endpoint's address (name).
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetTranslationName(ep_t ea, int i, en_t *gan)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	return ea->GetTranslationName(i, gan);
}

/*
 * AM_GetTranslationTag()
 * 
 * Purpose:	Fills in the tag of the endpoint mapped.
 * IN:		ea		-> The endpoint to search translation table of.
 *			i		-> The endpoint index number.
 * OUT:		tag		-> The remote endpoint tag.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetTranslationTag(ep_t ea, int i, tag_t *tag)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	*tag = ea->GetTranslationTag(i);
	return AM_OK;
}

/*
 * AM_GetTranslationInUse()
 * 
 * Purpose:	Returns AM_OK if the translation is in use.
 * IN:		ea		-> The endpoint to search translation table of.
 *			i		-> The endpoint index number.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetTranslationInUse(ep_t ea, int i)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	if(ea->GetTranslationInUse(i))
	{
		return AM_OK;
	}
	else
	{
		return AM_ERR_RESOURCE;
	}
}

/*
 * AM_GetNumTranslations()
 * 
 * Purpose:	Returns AM_OK if the translation is in use.
 * IN:		ea		-> The endpoint to search translation table of.
 * OUT:		ntrans	-> The size of the endpoint's translation table.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AM_GetNumTranslations(ep_t ea, int *ntrans)
{
	if(!ea)
	{
		return AM_ERR_BAD_ARG;
	}
	*ntrans = ea->GetNumTranslations();
	return AM_OK;
}


// ERROR STUFF!
void AM_SetErrorStream(ostream &errStream)
{
	gAMErrorStream = &errStream;
}

void AM_Error(char *s, int line, const char *file, int error_type)
{
#ifndef DEBUG_INFO
	if(error_type == AM_INFO)
	{
		return;
	}
#endif

	char	error[128];

	if(!gAMErrorStream) { return; }

	switch(error_type)
	{
	 case AM_OK:
		strcpy(error, "AM_OK");
		break;

	 case AM_ERR_NOT_INIT:
		strcpy(error, "AM_ERR_NOT_INIT");
		break;

	 case AM_ERR_BAD_ARG:
		strcpy(error, "AM_ERR_BAD_ARG");
		break;

	 case AM_ERR_RESOURCE:
		strcpy(error, "AM_ERR_RESOURCE");
		break;

	 case AM_ERR_NOT_SENT:
		strcpy(error, "AM_ERR_NOT_SENT");
		break;

	 case AM_ERR_IN_USE:
		strcpy(error, "AM_ERR_IN_USE");
		break;

	 case AM_INFO:
		 strcpy(error, "AM_INFO");
		 break;

	 case AM_EXCEPTION:
		 strcpy(error, "AM_EXCEPTION");
		 break;

	 case AM_ERROR:
		 strcpy(error, "ERROR");
		 break;
	}

	(*gAMErrorStream) << "[" << file << ":" << line << "] <" << error << ">: " << s << endl;	
	//	printf("[%s:%d] <%s>: %s\n", file, line, error, s);
	//	fflush(stdout);
}
