/* 
 * am.h
 *
 * Main header file for implementation of the Active Messages 2 interface.
 * See "Active Message Applications Programming Interface and Communication
 *  Subsystem Organization", Alan Mainwaring and David Culler.
 */
#ifndef __AM_H
#define __AM_H

#define	AM_MSG_WAIT_SPIN_COUNT	500

// Buffer Sizes (and other maximums)
#define AM_MAX_SHORT			8
#define AM_MAX_MEDIUM			512
#define AM_MAX_LONG				8192
#define AM_MAX_SEGMENT_SIZE		64*1024
#define AM_MAX_NUM_HANDLERS		255
#define AM_MAX_NUM_TRANSLATIONS	256

// Naming.
#define	AM_MAX_NODE_NAME_LENGTH 64	// Max length of a node's name.
#define AM_NAME_BASE			16	// Names are defined in hex!
#define AM_NAME_UNIQUE_SIZE		 8  // Size needed for 2^32 in hex + '\0'
#define AM_NAME_ENDPOINT_SIZE	 4	// Size needed for 2^16 in hex + '\0'
#define AM_NAME_BUNDLE_SIZE		 4	// Size needed for 2^16 in hex + '\0'

// Predefine some classes.
class cEndpoint;
class cBundle;
class cEndpointName;
class cAMLayer;
class cPacket;
class cConnection;

// Name Service Constants
#define AMNS_LOOKUP			0
#define AMNS_REGISTER		1
#define AMNS_UNREGISTER		2
#define AMNS_RESULT_ACK		0
#define AMNS_RESULT_NACK	1

/*
 *	--Interface Data Types--
 */
typedef int	EpName;
typedef struct
{
	char	node[AM_MAX_NODE_NAME_LENGTH];		// Node name: 
	char	unique[AM_NAME_UNIQUE_SIZE];		// Room for 2^32 in ascii
	char	bundle[AM_NAME_BUNDLE_SIZE];		// Room for 2^16 in ascii
	EpName	endpoint;							// An int for now.
} _en_t;

typedef unsigned int	handler_t;	// Handler table index.
typedef cEndpoint*		ep_t;		// Endpoint object.
typedef cBundle*		eb_t;		// Endpoint bundle object.
typedef _en_t*			en_t;		// Globally unique endpoint name.
typedef unsigned int	tag_t;		// Tag used for matching
typedef unsigned int	op_t;		// The opcode type for a returned AM.

typedef struct
{
	void	   *token;
	int			reply_endpoint;
	handler_t	handler;
	int			src_offset;
	int			dest_offset;
	void	   *source_addr;
	int			nbytes;
	int			args[AM_MAX_SHORT];
} _argblock_t;

typedef _argblock_t*	argblock_t;	// The argument block for a returned AM.


/*
 *	--Constants--
 */
#define AM_NONE 0	// Accept no messages (default).
#define	AM_ALL	1	// Accept all messages.

#define AM_PAR  0	// concurrent bundle/endpoint access.
#define AM_SEQ  1	// sequential bundle/endpoint access.

#define AM_OK			0
#define AM_ERR_NOT_INIT	1
#define AM_ERR_BAD_ARG	2
#define AM_ERR_RESOURCE	3
#define	AM_ERR_NOT_SENT	4
#define AM_ERR_IN_USE	5

#define AM_INFO			6
#define AM_EXCEPTION	7
#define AM_ERROR		8

/*
 *	--Handler Function Typedefs--
 */
typedef void (*Function)();
typedef void (*Undeliverable)(int status, op_t opcode, void *argblock);
typedef void (*Handler4)(void *token, int a0, int a1, int a2, int a3);
typedef void (*Handler8)(void *token, int a0, int a1, int a2, int a3, int a4, int a5, int a6, int a7);
typedef void (*Handler4I)(void *token, void *buf, int nbytes, int a0, int a1, int a2, int a3);
typedef void (*Handler8I)(void *token, void *buf, int nbytes, int a0, int a1, int a2, int a3, int a4, int a5, int a6, int a7);

/*
 *	--Return-To-Sender Rejection Error Codes--
 */
#define EBADARGS          1     /* Arguments to request or reply function invalid    */
#define EBADENTRY         2     /* X-lation table index selected unbound table entry */
#define EBADTAG           3     /* Sender's tag did not match the receiver's EP tag  */ 
#define EBADHANDLER       4     /* Invalid index into the recv.'s handler table      */ 
#define EBADSEGOFF        5     /* Offset into the dest-memory VM segment invalid    */
#define EBADLENGTH        6     /* Bulk xfer lenght goes beyond a segment's end      */
#define EBADENDPOINT      7     /* Destination endpoint does not exist               */
#define ECONGESTION       8     /* Congestion at destination endpoint                */
#define EUNREACHABLE      9     /* Destination endpoint unreachable                  */
#define EREPLYREJECTED    10    /* Destination endpoint refused reply message        */

/*
 *	--RTS Opcodes--
 */
#define AM_REQUEST_4					0
#define AM_REQUEST_8					1
#define AM_REQUEST_I4					2
#define AM_REQUEST_I8					3
#define AM_REPLY_4						4
#define AM_REPLY_8						5
#define AM_REPLY_I4						6
#define AM_REPLY_I8						7
#define AM_REQUEST_XFER4				8
#define AM_REPLY_XFER4					9

/*
 *	--Function Prototypes--
 */

/* Prints AM error strings. */
void AM_Error(char *s, int line, const char *file, int error_type=AM_ERROR);

/* Startup/Shutdown */
int	AM_Init();			//Initialize active message layer.
int	AM_Terminate();		//Shut down active message layer.


/* Request/Reply */
int AM_Request4(ep_t request_endpoint,			// Endpoint sending request.
				int reply_endpoint,				// Endpoint sending reply.
				handler_t handler,				// Index into dest handler table.
				int a0, int a1, int a2, int a3	// Args
				);
int AM_RequestI4(ep_t request_endpoint,			// Endpoint sending request.
				int reply_endpoint,				// Endpoint sending reply.
				handler_t handler,				// Index into dest handler table.
				void *source_addr,				// Source address of buf.
				int nbytes,						// Length of the source buffer.
				int a0, int a1, int a2, int a3	// Args
				);

int	AM_Reply4(void *token,						// ???
			  handler_t handler,				// Index into dest handler table.
			  int a0, int a1, int a2, int a3	// Args
			  );
int	AM_ReplyI4(void *token, handler_t handler, void *source_addr, int nbytes, int a0, int a1, int a2, int a3);

int AM_RequestXfer4(ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void *source_addr, int nbytes, int a0, int a1, int a2, int a3);
int AM_ReplyXfer4(void *token, int dest_offset, handler_t handler, void *source_addr, int nbytes, int a0, int a1, int a2, int a3);


int AM_GetSourceEndpoint(void *token,			// The token that was given in the handler.
						 en_t *gan				// OUT: The address is returned in this out param.
						 );
int	AM_GetDestEndpoint(void *token,				// The token that was given in the handler.
					   ep_t	*endp				// Returns the address of local-endpoint that received the token.
					   );
int AM_GetMsgTag(void *token, tag_t *tagp);

/* Endpoint/Bundle */
int	AM_AllocateBundle(int type,				// AM_SEQ | AM_PAR
					  eb_t *endb			// OUT: endb contains address of new bundle.
					  );

int	AM_AllocateEndpoint(eb_t bundle,		// Bundle to create endpoint in.
						ep_t *enp,			// OUT: enp contains address of new endpoint.
					    en_t *endpoint_name	// OUT: endpoint_name points to unique endpoint name.
						);

int AM_SetExpectedResources(ep_t ea, int n_endpoints, int n_outstanding_requests);

int	AM_Map(ep_t ea,			// IN: Endpoint to change remote address map table.
		   int index,		// IN: Index in map table to change.
		   en_t endpoint,	// IN: Remote endpoint to map to.
		   tag_t tag		// IN: The tag to be used when sending to endpoint.
		   );

int AM_MapAny(ep_t ea, int *index, en_t endpoint, tag_t tag);

int AM_Unmap(ep_t ea,		// IN: Endpoint to unmap entry in addr map table.
			 int index		// IN: Index in table to unmap.
			 );	

int	AM_FreeEndpoint(ep_t ea		// IN: Endpoint to free.
					);

int	AM_FreeBundle(eb_t bundle	// IN: Bundle to free.
				  );

/* Handler */
int	AM_SetHandler(ep_t ea,				// IN: The endpoint to set the handler for.
				  handler_t handler,	// IN: The index into the ep's handler table.
				  void (*function)()	// IN: Pointer to handler function.
				  );
int AM_SetHandlerAny(ep_t ea, handler_t *handler, void (*function)());
int AM_GetNumHandlers(ep_t ea, int *n_handlers);
int AM_SetNumHandlers(ep_t ea, int n_handlers);
int AM_MaxNumHandlers();

/* Segment */
int AM_SetSeg(ep_t ea, void *addr, int nbytes);
int AM_GetSeg(ep_t ea, void **addr, int *nbytes);
int AM_MaxSegLength(int *nbytes);

/* Misc */
int	AM_Poll(eb_t bundle);				// IN: The bundle to poll on.
int AM_MaxShort();
int AM_MaxMedium();
int AM_MaxLong();

/* Translations/Tags */
int AM_SetTag(ep_t ea, tag_t tag);
int AM_GetTag(ep_t ea, tag_t *tag);
int AM_GetTranslationName(ep_t ea, int i, en_t *gan);
int AM_GetTranslationTag(ep_t ea, int i, tag_t *tag);
int AM_GetTranslationInUse(ep_t ea, int i);
int AM_MaxNumTranslations(int *ntrans);
int AM_GetNumTranslations(ep_t ea, int *ntrans);
//int AM_SetNumTranslations(ep_t ea, int ntrans);


// NAME SERVICE
int AMNS_Init(char *nameServer);	// Input: The node where the name server is on.
int AMNS_Lookup(char *name, _en_t *result);
int AMNS_Register(char *name, en_t epn);
int AMNS_Unregister(char *name);
#endif