/* 
 * amns.h
 *
 * Implementation of most of the AM name service functions.
 */
#include "amns.h"

char _AMNS_ServerName[128];
bool _AMNS_Initialized = false;

/*
 * AMNS_Init()
 * 
 * Purpose:	Sets the node name where the name service can be found.
 * IN:		name	-> The node where the name service is located.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The name service is initialized.
 * Return:	AM_OK or AM_ERR_*
 */
int AMNS_Init(char *nameServer)
{
	strcpy(_AMNS_ServerName, nameServer);
	_AMNS_Initialized = true;
	return AM_OK;
}

/*
 * AMNS_Lookup()
 * 
 * Purpose:	Does a name service lookup based on the given string.
 * IN:		name	-> The name to look up.
 *			result	-> The endpointname that is filled in if success.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AMNS_Lookup(char *name, _en_t *result)
{
	cNic	*nic;
	cDesc	*outDesc;
	cDesc	*inDesc;
	cDSeg	*outseg1;
	cDSeg	*inseg;
	int		 retVal;

	if (!gAMLayer || !_AMNS_Initialized)
	{
		return AM_ERR_NOT_INIT;
	}
	nic = gAMLayer->GetNic();
	outDesc = nic->CreateDesc(2, gAMLayer->GetPtag());
	if (!outDesc) { return AM_ERR_RESOURCE; }
	inDesc  = nic->CreateDesc(2, gAMLayer->GetPtag());
	if (!inDesc)
	{
		delete outDesc;
		return AM_ERR_RESOURCE;
	}
	outseg1 = nic->CreateDSeg(NS_MAX_NAME_SIZE, gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!outseg1)
	{
		return AM_ERR_RESOURCE;
	}
	inseg = nic->CreateDSeg(sizeof(_en_t), gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!inseg)
	{
		return AM_ERR_RESOURCE;
	}

	// Set up the out descriptor.
	outDesc->SetControl(VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE);
	outDesc->SetSegCount(1);
	outDesc->SetDSeg(0, outseg1);
	outDesc->SetLength(NS_MAX_NAME_SIZE);

	// Set up the in descriptor.
	inDesc->SetControl(VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE);
	inDesc->SetSegCount(1);
	inDesc->SetDSeg(0, inseg);
	inDesc->SetLength(sizeof(_en_t));

	// Set up the outgoing message.
	memset(outseg1->GetData(), 0, NS_MAX_NAME_SIZE);
	strcpy((char *)outseg1->GetData(), name);
	outDesc->SetImmediateData(NS_LOOKUP);

	if(_AMNS_SendRecv(outDesc, inDesc))
	{
		if(inDesc->GetImmediateData() == NS_ACK)
		{
			memcpy(result, inseg->GetData(), sizeof(_en_t));
			retVal = AM_OK;
		}
		else
		{
			retVal = AM_ERR_RESOURCE;
		}
	}
	else
	{
		retVal = AM_ERR_RESOURCE;
	}

	// Cleanup
	delete outseg1;
	delete inseg;
	delete outDesc;
	delete inDesc;

	return retVal;
}

/*
 * AMNS_Register()
 * 
 * Purpose:	Register the given endpoint addr under name.
 * IN:		name	-> The name to register under.
 *			epn		-> The endpointname to register.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AMNS_Register(char *name, en_t epn)
{
	cNic	*nic;
	cDesc	*outDesc;
	cDesc	*inDesc;
	cDSeg	*outseg1;
	cDSeg	*outseg2;
	cDSeg	*inseg;
	int		 retVal;

	if (!gAMLayer || !_AMNS_Initialized)
	{
		return AM_ERR_NOT_INIT;
	}
	nic = gAMLayer->GetNic();
	outDesc = nic->CreateDesc(2, gAMLayer->GetPtag());
	if (!outDesc) { return AM_ERR_RESOURCE; }
	inDesc  = nic->CreateDesc(2, gAMLayer->GetPtag());
	if (!inDesc)
	{
		delete outDesc;
		return AM_ERR_RESOURCE;
	}
	outseg1 = nic->CreateDSeg(NS_MAX_NAME_SIZE, gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!outseg1)
	{
		return AM_ERR_RESOURCE;
	}
	outseg2 = nic->CreateDSeg(sizeof(_en_t), gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!outseg2)
	{
		return AM_ERR_RESOURCE;
	}
	inseg = nic->CreateDSeg(sizeof(_en_t), gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!inseg)
	{
		return AM_ERR_RESOURCE;
	}

	// Set up the out descriptor.
	outDesc->SetControl(VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE);
	outDesc->SetSegCount(2);
	outDesc->SetDSeg(0, outseg1);
	outDesc->SetDSeg(1, outseg2);
	outDesc->SetLength(NS_MAX_NAME_SIZE+sizeof(_en_t));

	// Set up the in descriptor.
	inDesc->SetControl(VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE);
	inDesc->SetSegCount(1);
	inDesc->SetDSeg(0, inseg);
	inDesc->SetLength(sizeof(_en_t));

	// Set up the outgoing message.
	memset(outseg1->GetData(), 0, NS_MAX_NAME_SIZE);
	strcpy((char *)outseg1->GetData(), name);
	memcpy(outseg2->GetData(), epn, sizeof(_en_t));
	outDesc->SetImmediateData(NS_REGISTER);

	if(_AMNS_SendRecv(outDesc, inDesc))
	{
		if(inDesc->GetImmediateData() == NS_ACK)
		{
			retVal = AM_OK;
		}
		else
		{
			retVal = AM_ERR_RESOURCE;
		}
	}
	else
	{
		retVal = AM_ERR_RESOURCE;
	}

	// Cleanup
	delete outseg1;
	delete outseg2;
	delete inseg;
	delete outDesc;
	delete inDesc;

	return retVal;
}

/*
 * AMNS_Unregister()
 * 
 * Purpose:	Unregisters the entry corresponding to the given name.
 * IN:		name	-> The name to unregister.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int AMNS_Unregister(char *name)
{
	cNic	*nic;
	cDesc	*outDesc;
	cDesc	*inDesc;
	cDSeg	*outseg1;
	cDSeg	*inseg;
	int		 retVal;

	if (!gAMLayer || !_AMNS_Initialized)
	{
		return AM_ERR_NOT_INIT;
	}
	nic = gAMLayer->GetNic();
	outDesc = nic->CreateDesc(2, gAMLayer->GetPtag());
	if (!outDesc) { return AM_ERR_RESOURCE; }
	inDesc  = nic->CreateDesc(2, gAMLayer->GetPtag());
	if (!inDesc)
	{
		delete outDesc;
		return AM_ERR_RESOURCE;
	}
	outseg1 = nic->CreateDSeg(NS_MAX_NAME_SIZE, gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!outseg1)
	{
		return AM_ERR_RESOURCE;
	}
	inseg = nic->CreateDSeg(sizeof(_en_t), gAMLayer->GetPtag(), VIP_FALSE, VIP_FALSE);
	if(!inseg)
	{
		return AM_ERR_RESOURCE;
	}

	// Set up the out descriptor.
	outDesc->SetControl(VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE);
	outDesc->SetSegCount(1);
	outDesc->SetDSeg(0, outseg1);
	outDesc->SetLength(NS_MAX_NAME_SIZE);

	// Set up the in descriptor.
	inDesc->SetControl(VIP_CONTROL_OP_SENDRECV | VIP_CONTROL_IMMEDIATE);
	inDesc->SetSegCount(1);
	inDesc->SetDSeg(0, inseg);
	inDesc->SetLength(sizeof(_en_t));

	// Set up the outgoing message.
	memset(outseg1->GetData(), 0, NS_MAX_NAME_SIZE);
	strcpy((char *)outseg1->GetData(), name);
	outDesc->SetImmediateData(NS_UNREGISTER);

	if(_AMNS_SendRecv(outDesc, inDesc))
	{
		if(inDesc->GetImmediateData() == NS_ACK)
		{
			retVal = AM_OK;
		}
		else
		{
			retVal = AM_ERR_RESOURCE;
		}
	}
	else
	{
		retVal = AM_ERR_RESOURCE;
	}

	// Cleanup
	delete outseg1;
	delete inseg;
	delete outDesc;
	delete inDesc;

	return retVal;
}


/*
 * _AMNS_SendRecv()
 * 
 * Purpose:	Connects to the name service and exchanges information.
 * IN:		outDesc	-> The descriptor to send to ns.
 *			inDesc	-> The descriptor to receive from ns.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true if success, else false.
 */
bool _AMNS_SendRecv(cDesc *outDesc, cDesc *inDesc)
{
	VIP_RETURN	 rc;
	cVI			*vi;
	cVIAddr		*localAddr;
	cVIAddr		*remoteAddr;
	cNic		*nic;
	VIP_VI_ATTRIBUTES viAttrib;

	nic = gAMLayer->GetNic();
	const VIP_NIC_ATTRIBUTES *nicAttrib;
	nicAttrib = nic->GetAttrib();
	viAttrib.ReliabilityLevel = VIP_SERVICE_RELIABLE_DELIVERY;
	viAttrib.MaxTransferSize  = nicAttrib->MaxTransferSize;
	viAttrib.QoS			  = 0;
	viAttrib.Ptag			  = gAMLayer->GetPtag();
	viAttrib.EnableRdmaWrite  = VIP_FALSE;
	viAttrib.EnableRdmaRead   = VIP_FALSE;
	vi = nic->CreateVI(&viAttrib);
	if(!vi)
	{
		return NULL;
	}

	// Create addresses.
	localAddr = nic->CreateVIAddr();
	localAddr->GetLocalAddress();
	remoteAddr = nic->CreateVIAddr();
	remoteAddr->GetHostByName(_AMNS_ServerName);
	remoteAddr->SetDiscriminator("ns", 2);
	
	
	rc = vi->PostRecv(inDesc);
	if(rc != VIP_SUCCESS)
	{
		printf("could not post recv.\n");
		return false;
	}
	rc = vi->ConnectRequest(localAddr, remoteAddr, 5000);
	if(rc != VIP_SUCCESS)
	{
		printf("Could not connect to ns.\n");
		return false;
	}
	vi->PostSend(outDesc);
	if(rc != VIP_SUCCESS)
	{
		printf("Could not send to ns.\n");
		return false;
	}
	vi->SendWait(outDesc, 3000);
	if(rc != VIP_SUCCESS)
	{
		printf("Could not wait for send.\n");
		return false;
	}
	vi->RecvWait(inDesc, 3000);
	if(rc != VIP_SUCCESS)
	{
		printf("Could not recv.\n");
		return false;
	}

	vi->Disconnect();

	return true;
}