/*
 * cAMLayer.cpp
 * 
 * Class implementatino for AM layer class. 
 *
 * Authors:  Vicky & Ted
 */
#include "cAMLayer.h"
#include <process.h>
#include "cBundle.h"
#include "cHashTable.h"
#include "cConnection.h"

cHashTable *gAMConnectionTable = NULL;

/*
 * cAMLayer::cAMLayer()
 *
 * Purpose:	Create a new Active Messages layer.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The AM's layer is created but not initialized.
 * Return:	-
 */
cAMLayer::cAMLayer()
{
	mNic			 = NULL;
	mPtag			 = NULL;
	mBundleArray	 = NULL;		
	mLocalName		 = NULL;
	gAMConnectionTable = NULL;
}

void ErrorHandler(VIP_PVOID context, VIP_ERROR_DESCRIPTOR *errorDesc)
{
	VIP_RETURN	rc;
	VIP_VI_HANDLE			mVI;
	VIP_VI_ATTRIBUTES		mAttribs;
	VIP_VI_STATE			mState;
	VIP_BOOLEAN				mSendQEmpty;
	VIP_BOOLEAN				mRecvQEmpty;
	
	cConnection			   *connection;

	printf("\n\n!!!!!!Error handler!!!!!!\n");
	cVIError::SetErrorStream(cerr);
	cVIError::PrintError("Inside error handler",  __LINE__, __FILE__, errorDesc->ErrorCode);

	// Query the VI
	mVI = errorDesc->ViHandle;
	rc = VipQueryVi(mVI, &mState, &mAttribs, &mSendQEmpty, &mRecvQEmpty);
	if(rc != VIP_SUCCESS)
	{
		cVIError::PrintError("In error handler, could not query VI!", __LINE__, __FILE__, rc);
	}
	printf("State: %d\n", mState);

	if(gAMConnectionTable->Find(mVI, (void **)&connection))
	{
//		printf("Connection: %x\n", connection);

		// Handle asynchronous error.
		switch(errorDesc->ErrorCode)
		{

		case VIP_ERROR_CONN_LOST:
			if(connection->mNode)
			{
				// Incoming connection -- Close it.
				printf("Closing incoming connection [%x]\n", connection);
				connection->CloseConnection();
			}
			else
			{
				// Outgoing connection -- Merely disable it.
				printf("Disabling outgoing connection [%x]\n", connection);
				connection->DisableConnection();
			}
			printf("\n\n");
			fflush(stdout);
			break;

		case VIP_ERROR_RECVQ_EMPTY:
			connection->RecvQEmptyWarning();
			break;

		default:
			AM_Error("Uhandled asynchronous error.", __LINE__, __FILE__);
		} /* switch */
	} /* if connection */
}

/*
 * cAMLayer::Initialize()
 *
 * Purpose:	Initialize the AM layer for first use.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The AM layer is initialized.
 * Return:	true if success or false otherwise.
 */
bool cAMLayer::Initialize(char *deviceName)
{
	char			localName[255];
	unsigned long	nameLen = AM_MAX_NODE_NAME_LENGTH;


	// Get the local node name (first part of address)
	if(!cVIAddr::GetLocalNodeName(localName, &nameLen))
	{
		return false;
	}
	else
	{
		mLocalName = new char[nameLen+1];
		strcpy(mLocalName, localName);
	}

	// Get the "unique" number (second part of address)
	mUnique = getpid();

	// Do VI initialization stuff.
	mNic = new cNic(deviceName);
	if(!mNic)
	{
		return false;
	}
	if(!(mNic->IsOpen()))
	{
		delete mNic;
		mNic = NULL;
		return false;
	}
	if(mNic->ErrorCallback(NULL, ErrorHandler) != VIP_SUCCESS)
	{
		return false;
	}
	mPtag = mNic->CreatePtag();
	if(!mPtag)
	{
		delete mNic;
		mNic = NULL;
		return false;
	}

	// Create some sort of data structure to hold bundles.
	mBundleArray = new cBundle*[DEFAULT_BUNDLE_ARRAY_SIZE];
	if(!mBundleArray)
	{
		return false;
	}
	AM_Error("Clearing the bundle array in init.", __LINE__, __FILE__, AM_INFO);
	for(int i = 0; i < DEFAULT_BUNDLE_ARRAY_SIZE; i++)
	{
		mBundleArray[i] = NULL;
	}

	// Create the connection table.
	gAMConnectionTable = new cHashTable(32);	// CHANGE FROM CONSTANT!!
	if(!gAMConnectionTable)
	{
		return false;
	}

	// Success!
	return true;
}

/*
 * cAMLayer::~cAMLayer()
 *
 * Purpose:	Destroy the Active Messages layer.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The endpoints resources are de-allocated and destroyed.
 * Return:	-
 */
cAMLayer::~cAMLayer()
{

	AM_Error("Deleting AM Layer.", __LINE__, __FILE__, AM_INFO);

	if(mBundleArray)
	{
		AM_Error("Cleaning up bundles in layer.", __LINE__, __FILE__, AM_INFO);
		//Free all of the bundles (and their endpoints).
		for(int i = 0; i < DEFAULT_BUNDLE_ARRAY_SIZE; i++)
		{	
			if(mBundleArray[i] != NULL)
			{
				AM_Error("Freeing an endpoint.", __LINE__, __FILE__, AM_INFO);
				FreeBundle(mBundleArray[i]);
			}
		}
		AM_Error("Done cleaning up bundles in layer.", __LINE__, __FILE__, AM_INFO);
		delete[] mBundleArray;
	}

	//Close and delete the Nic
	if(mNic)
	{
		AM_Error("Closing the Nic.", __LINE__, __FILE__, AM_INFO);
		delete mNic;
	}

	// Delete the connection table.
	if(gAMConnectionTable)
	{
		delete gAMConnectionTable;
	}
}

/*
 * cAMLayer::AllocateBundle()
 *
 * Purpose:	Allocates a bundle and adds it to the set of bundles for this layer.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The bundle is added to the AMLayer's bundle list.
 * Return:	NULL if fail, or a cBundle pointer.
 */
cBundle* cAMLayer::AllocateBundle()
{
	cBundle		*newBundle;


	// Find an empty slot.
	for(int i = 0; i < DEFAULT_BUNDLE_ARRAY_SIZE; i++)
	{
		if(mBundleArray[i] == NULL)
		{
			// Attempt to allocate a bundle here.
			newBundle = new cBundle(this, i);
			if(!newBundle)
			{
				return NULL;
			}
			if(!newBundle->Init())
			{
				AM_Error("AllocateBundle(): Unable to init new bundle.", __LINE__, __FILE__);
				delete newBundle;
				return NULL;
			}
			mBundleArray[i] = newBundle;
			if(!(newBundle->Start()))
			{
				AM_Error("AllocateBundle(): Unable to start new bundle's thread.", __LINE__, __FILE__);
				delete newBundle;
				return NULL;
			}
			return newBundle;
		}
	}
	return false;
}

/*
 * cAMLayer::FreeBundle()
 *
 * Purpose:	Removes the bundle from the set of bundles for this AMLayer.
 *			And deletes it.
 * IN:		bundle	-> Pointer to the bundle to be trashed.
 * OUT:		-
 * Cond:	The bundle must be in the AMLayer's bundle list.
 * PostCnd:	The bundle is removed from the AMLayer's bundle list.
 * Return:	AM_OK if success, AM_ERR_* otherwise.
 */
int cAMLayer::FreeBundle(cBundle *bundle)
{
	int	index;

	index = bundle->GetNumber();
	if(mBundleArray[index] == NULL)
	{
		AM_Error("FreeBundle(): Bundle does not exist in mBundleArray.", __LINE__, __FILE__, AM_ERR_BAD_ARG);
		return AM_ERR_BAD_ARG;
	}
	mBundleArray[index] = NULL;
	delete bundle;
	return AM_OK;
}

/*
 * cAMLayer::GetNic()
 *
 * Purpose:	mNic accessor function.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	A pointer to the cNic object.
 */
cNic* cAMLayer::GetNic()
{
	return mNic;
}

char* cAMLayer::GetNodeName()
{
	return mLocalName;
}
int cAMLayer::GetNodeNameLen()
{
	return mNameLen;
}
int cAMLayer::GetUnique()
{
	return mUnique;
}