/*
 * cBundle.cpp
 * 
 * Class implementatino for AM bundle class. 
 *
 * Authors:  Vicky & Ted
 * Last Mod: 4 March 1999 by ted
 */
#include "cBundle.h"

#include "cAMLayer.h"
#include "cPacket.h"
#include "cEndpoint.h"
#include "cConnection.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>

DWORD WINAPI BundleThread(LPVOID lpParameter);
DWORD WINAPI ConnectThread(LPVOID lpParameter);

/*
 * cBundle::cBundle()
 *
 * Purpose:	Create and initialize a new bundle.  Does not start thread.
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer is initialized.
 * PostCnd:	The bundle's data structures are allocated and intialized.
 * Return:	-
 */
cBundle::cBundle(cAMLayer *layer, int number)
{
	mPtag = NULL;
	mCQ	  = NULL;

	mRunMain = false;
	mRunConnection = false;
	mNumber	 = number;
	mAMLayer = layer;

	mEndpointArray		= NULL;		
	mLocalAddress		= NULL;		
	mUsedConnectionQ	= NULL;
	mClosedConnectionQ	= NULL;
	mNic = layer->GetNic();
}

/*
 * cBundle::Init()
 *
 * Purpose:	Create and intializes the new bundle.  
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer is initialized.
 * PostCnd:	The bundle's data structures are allocated and intialized.
 * Return:	true if success, else false.
 */
bool cBundle::Init()
{
	VIP_BOOLEAN	retVal;
	char		temp[64];

	// Fill in the bundle name structure.
	memset(&mBundleAMAddr, 0, sizeof(_en_t));
	strcpy(mBundleAMAddr.node, mAMLayer->GetNodeName());

	memset(temp, 0, sizeof(temp));
	_itoa(mNumber, temp, AM_NAME_BASE);
	memcpy(mBundleAMAddr.bundle, temp, AM_NAME_BUNDLE_SIZE);
	
	memset(temp, 0, sizeof(temp));
	_itoa(mAMLayer->GetUnique(), temp, AM_NAME_BASE);
	memcpy(mBundleAMAddr.unique, temp, AM_NAME_UNIQUE_SIZE);

	// Make the endpoint array.
	mEndpointArraySize = DEFAULT_ENDPOINT_ARRAY_SIZE;
	mEndpointArray = new cEndpoint*[DEFAULT_ENDPOINT_ARRAY_SIZE];
	if(!mEndpointArray)
	{
		return false;
	}
	for(int i = 0; i < mEndpointArraySize; i++)
	{
		mEndpointArray[i] = NULL;
	}

	// Create the Ptag
	mPtag = (mAMLayer->GetNic())->CreatePtag();
	if(!mPtag)	
	{ 
		return false; 
	}

	// Make the local address.
	mLocalAddress = (mAMLayer->GetNic())->CreateVIAddr();
	if(!mLocalAddress)
	{
		return false;
	}

	retVal = mLocalAddress->GetLocalAddress();
	if (retVal == VIP_FALSE)
	{
		return false;
	}

	// Set up the discriminator.
	retVal = mLocalAddress->SetDiscriminator(mBundleAMAddr.unique, sizeof(AM_NAME_UNIQUE_SIZE+AM_NAME_BUNDLE_SIZE));
	if (retVal == VIP_FALSE)
	{
		return false;
	}

	// Create the completion queue.
	mCQ = (mAMLayer->GetNic())->CreateCQ(128);
	if (!mCQ)
	{
		return false;
	}

	// Create the resource fifos.
	try
	{
		mUsedConnectionQ	= new cFifo(32);
		mClosedConnectionQ  = new cFifo(32);
	}
	catch(cFifoException &e)
	{
		AM_Error(e.ToString(), __LINE__, __FILE__, AM_EXCEPTION);
		return false;
	}
	mAcks			 = 0;

	// Verify the allocations.
	if(!mCQ || !mUsedConnectionQ || !mClosedConnectionQ)
	{
		return false;
	}
	return true;
}

/*
 * cBundle::~cBundle()
 *
 * Purpose:	Destroy a bundle.  Clean up all data structures.
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer is initialized.
 * PostCnd:	The bundle's data structures are cleaned up and de-allocated.
 * Return:	-
 */
cBundle::~cBundle()
{

	AM_Error("~cBundle(): Destroying this bundle.", __LINE__, __FILE__, AM_INFO);

	// If still running, then stop.
	if(mRunConnection)
	{
		_StopConnect();
	}

	// Delete connections.
	void	*temp;
	cConnection *tempConn;

	// Free the used connections.
	AM_Error("~cBundle(): Deleting used connection queue.", __LINE__, __FILE__, AM_INFO);
	if(mUsedConnectionQ)
	{
		while(mUsedConnectionQ->Peek(&temp) != NULL)
		{
			tempConn = (cConnection *)temp;
			assert(tempConn);

			if(!tempConn->IsActive())
			{
				// Hack - If this is a pending connection, just delete it.
				mUsedConnectionQ->Pop((void **)&tempConn);
				delete tempConn;
			}
			else
			{
				// Close down and delete properly.
				tempConn->CloseConnection();
			}
		}
		delete mUsedConnectionQ;
	}

	// Clean out the endpoint array.
	if(mEndpointArray)
	{
		// Free all left-over endpoints.
		for(int i=0; i<mEndpointArraySize; i++)
		{
			if(mEndpointArray[i] != NULL)
			{
				AM_Error("~cBundle(): Freeing endpoint.", __LINE__, __FILE__, AM_INFO);
				FreeEndpoint(mEndpointArray[i]);
			}
		}
		AM_Error("~cBundle(): Deleting endpoint array.", __LINE__, __FILE__, AM_INFO);
		delete[] mEndpointArray;
	}

	if(mRunMain)
	{
		_StopMain();
	}

	// Get rid of the local address object.
	if(mLocalAddress)
	{
		AM_Error("~cBundle(): Deleting local address.", __LINE__, __FILE__, AM_INFO);
		delete mLocalAddress;
	}

	// Delete the completion queue.
	AM_Error("~cBundle(): Deleting completion queue.", __LINE__, __FILE__, AM_INFO);
	if(mCQ)
	{
		delete mCQ;
	}

	// Delete the ptag.
	AM_Error("~cBundle(): ptag?", __LINE__, __FILE__, AM_INFO);
	if(mPtag)
	{
		delete mPtag;
	}
	AM_Error("~cBundle(): Done w/ cleanup.", __LINE__, __FILE__, AM_INFO);

	// Free the closed connections.
	AM_Error("~cBundle(): Deleting closed connection queue.", __LINE__, __FILE__, AM_INFO);
	if(mClosedConnectionQ)
	{
		while(mClosedConnectionQ->Pop(&temp))
		{
			tempConn = (cConnection *)temp;
			assert(tempConn);
			delete tempConn;
		}
		delete mClosedConnectionQ;
	}
}

/*
 * cBundle::Start()
 *
 * Purpose:	Creates a new thread for this bundle and starts execution.
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer is initialized.
 * PostCnd:	A new thread is initialized and running to handle AM's to/from this bundle.
 * Return:	true if OK, else false.  
 */
bool cBundle::Start()
{
	// Make sure it is not already started.
	if (mRunMain && mRunConnection)
	{
		return false;
	}

	// Thread settings vars
	LPSECURITY_ATTRIBUTES	lpThreadAttributes;
	DWORD					dwStackSize;
	LPTHREAD_START_ROUTINE	lpStartAddress;
	LPVOID					lpParameter;
	DWORD					dwCreationFlags;
	LPDWORD					lpThreadId;

	// Allow threads to run
	mRunMain = true;
	mRunConnection = true;

	// Set up the main thread settings.
	lpThreadId			= &mMainThreadId;
	dwCreationFlags		= NULL;
	lpParameter			= this;	//Address of this object.		
	dwStackSize			= 0;	//Same as starting parent.
	lpStartAddress		= BundleThread;
	lpThreadAttributes	= NULL;

	// Start the main thread.
	mMainThreadHandle = CreateThread(
		lpThreadAttributes,  // pointer to security attributes
		dwStackSize,         // initial thread stack size
		lpStartAddress,      // pointer to thread function
		lpParameter,         // argument for new thread
		dwCreationFlags,     // creation flags
		lpThreadId           // pointer to receive thread ID
	);
	if(mMainThreadHandle == NULL)
	{
		mRunMain = false;
		return mRunMain;
	}

	// Set up the connection thread settings.
	lpThreadId			= &mConnectThreadId;
	dwCreationFlags		= NULL;
	lpParameter			= this;	//Address of this object.		
	dwStackSize			= 0;	//Same as starting parent.
	lpStartAddress		= ConnectThread;
	lpThreadAttributes	= NULL;

	// Start the connection thread.
	mConnectThreadHandle = CreateThread(
		lpThreadAttributes,  // pointer to security attributes
		dwStackSize,         // initial thread stack size
		lpStartAddress,      // pointer to thread function
		lpParameter,         // argument for new thread
		dwCreationFlags,     // creation flags
		lpThreadId           // pointer to receive thread ID
	);
	if(mConnectThreadHandle == NULL)
	{
		_StopMain();
		mRunConnection = false;
		return mRunConnection;
	}
	return (mRunMain && mRunConnection);
}

/*
 * cBundle::Stop()
 *
 * Purpose:	Terminates the current thread for this bundle and re-inits all data structures.
 * IN:		-
 * OUT:		-
 * Cond:	The AM Layer is initialized.
 * PostCnd:	The thread for this bundle is stopped.  All endpoints in the bundle are dropped
 *			and all bundle data structures are re-initialized as if the bundle has been created
 *			anew: The next call to Start should startup the bundle fresh.
 * Return:	true if OK, else false.  
 */
bool cBundle::Stop()
{
	bool	ret1, ret2;
	ret1 = _StopMain();
	ret2 = _StopConnect();
	return (ret1 && ret2);
}
bool cBundle::_StopMain()
{
	DWORD	retCode = STILL_ACTIVE;

	mRunMain = false;

	while(retCode == STILL_ACTIVE)
	{
		Sleep(3);
		GetExitCodeThread(mMainThreadHandle, &retCode);
	}
	AM_Error("Stop(): Main Thread done.", __LINE__, __FILE__, AM_INFO);
	return true;
}
bool cBundle::_StopConnect()
{
	DWORD	retCode = STILL_ACTIVE;

	mRunConnection = false;
	while(retCode == STILL_ACTIVE)
	{
		Sleep(3);
		GetExitCodeThread(mConnectThreadHandle, &retCode);
	}
	AM_Error("Stop(): Connect Thread done.", __LINE__, __FILE__, AM_INFO);
	return true;
}
bool cBundle::IsRunning()
{
 	return (mRunMain && mRunConnection);
}

/*
 * cBundle::AllocateEndpoint()
 *
 * Purpose:	Create a new endpoint within this bundle.
 * IN:		-
 * OUT:		ep			-> Filled with address of the newly allocated endpoint.
 *			endpontName -> Set to the address of the new endpoint's address.
 * Cond:	Available resources.
 * PostCnd:	A new endpoint is added to this bundle and returned in ep.
 * Return:	AM_OK or AM_ERR_*
 */
int cBundle::AllocateEndpoint(cEndpoint **ep, en_t *endpointName)
{
	int		endpointNumber = mEndpointArraySize + 1;

	mCSEndpointArray.Enter();

	// Find spot in array to put endpoint.
	for(int i=0; i<mEndpointArraySize; i++)
	{
		if(mEndpointArray[i] == NULL)
		{
			endpointNumber = i;
			break;
		}
	}
	if(endpointNumber > mEndpointArraySize)
	{
		mCSEndpointArray.Leave();
		return AM_ERR_RESOURCE;
	}

	// Alloc endpoint and put it into array.
	try
	{
		*ep = new cEndpoint(this, endpointNumber);
		if(!*ep)
		{
			mCSEndpointArray.Leave();
			return AM_ERR_RESOURCE;
		}
	}
	catch(AMException &e)
	{
		AM_Error(e.ToString(), __LINE__, __FILE__, AM_EXCEPTION);
		delete *ep;
		*ep = NULL;
		mCSEndpointArray.Leave();
		return AM_ERR_RESOURCE;
	}
	mEndpointArray[endpointNumber] = *ep;

	mCSEndpointArray.Leave();
	
	// Fill in the name structure.
	*endpointName = (*ep)->GetAddress();
	mBundleAMAddr.endpoint = endpointNumber;
	(*ep)->SetAddress(&mBundleAMAddr);

	// **DEBUG**
//	printf("New Endpoint:%s:%s:%s:%d\n", (*endpointName)->node, (*endpointName)->unique, (*endpointName)->bundle, (*endpointName)->endpoint);
	// ** END DEBUG **

	return AM_OK;
}

/*
 * cBundle::FreeEndpoint()
 *
 * Purpose:	Frees the given endpoint if it is present.
 * IN:		ep	-> Pointer to endpoint to be freed.
 * OUT:		-
 * Cond:	ep is a pointer to a valid endpoint.
 * PostCnd:	The endpoint is deleted and removed from the ep array.
 * Return:	AM_OK or AM_ERR_*
 */
int	cBundle::FreeEndpoint(cEndpoint *ep)
{
	int	epNumber;

	mCSEndpointArray.Enter();

		epNumber = ep->GetNumber();
		mEndpointArray[epNumber] = NULL;
		delete	ep;

	mCSEndpointArray.Leave();
	return AM_OK;
}

/*
 * cBundle::_BundleMain()
 *
 * Purpose:	The thread entry point for the bundle.
 * IN:		-
 * OUT:		-
 * Cond:	A thread has been created w/ entry point BundleThread.
 * PostCnd:	_BundleMain() will loop until the thread is terminated.
 * Return:	Proper thread exit return code.
 */
DWORD cBundle::_BundleMain()
{
	DWORD			 retCode = 0;
	VIP_RETURN		 rc;
	VIP_VI_HANDLE	 viHandle;
	VIP_BOOLEAN		 recvQ;
	VIP_DESCRIPTOR	*DescPtr;
	Header			*header;
	cPacket			*packet;
	cConnection		*connection;

	while(mRunMain)
	{
		// Wait on CQ
		for(int i = 0; i < AM_MSG_WAIT_SPIN_COUNT; i++)
		{
			rc = mCQ->Done(&viHandle, &recvQ);
			if(rc != VIP_NOT_DONE) { break; }
		}
		if(rc == VIP_NOT_DONE)
		{
			rc = mCQ->Wait(2000, &viHandle, &recvQ);
		}
		if(rc == VIP_SUCCESS)
		{

			// Recv Queue.
			if(recvQ == VIP_TRUE)
			{
				rc = cVI::RecvDone(viHandle, &DescPtr);
				if(rc == VIP_SUCCESS)
				{
					// RECV SUCCESS
					header = (Header *)cDesc::GetData(DescPtr, 0);
					connection = header->hidden.connPtr;

					// Handle the receive.
					packet = (cPacket *)header->hidden.packetPtr;
					connection->RecvPacket(packet);
				}
				else
				{
					if(DescPtr == NULL)
					{
						AM_Error("Pulled NULL descriptor off of recv queue!", __LINE__, __FILE__);
					}
					else if(rc != VIP_DESCRIPTOR_ERROR)
					{
						cVIError::PrintError("Cannot pull desc. from recv queue.", __LINE__, __FILE__, rc);
					}
					else
					{
						// Descriptor completed in error.  Return free packet to queue.
						header = (Header *)cDesc::GetData(DescPtr, 0);
						packet = (cPacket *)header->hidden.packetPtr;
						connection = (cConnection *)header->hidden.connPtr;				
						connection->ReleasePacket(packet);
					}
				}
			}	
			
			// Send Queue.
			else
			{
				rc = cVI::SendDone(viHandle, &DescPtr);
				if(rc == VIP_SUCCESS)
				{
					header		= (Header *)cDesc::GetData(DescPtr, 0);
					packet		= (cPacket *)header->hidden.packetPtr;
					connection  = (cConnection *)header->hidden.connPtr;
					_HandleSend(connection, packet);
				}
				else
				{
					if(DescPtr == NULL)
					{
						AM_Error("Pulled NULL descriptor off of send queue!", __LINE__, __FILE__);
					}
					else if(rc != VIP_DESCRIPTOR_ERROR)
					{
						cVIError::PrintError("Cannot pull desc. from send queue.", __LINE__, __FILE__, rc);
					}
					else
					{
						// Don't want to free descriptor 'cause it should be in a send window somewhere!
						header = (Header *)cDesc::GetData(DescPtr, 0);
						packet = (cPacket *)header->hidden.packetPtr;
						connection = (cConnection *)header->hidden.connPtr;
						AM_Error("Send descriptor completed in error.", __LINE__, __FILE__);
						_HandleSend(connection, packet);	// If it was an ack, then it should be handled normally and lost!
					}
					continue;
				}
			}
		} // end cq success.
	} // end while(mRun)

	while(mCQ->Done(&viHandle, &recvQ) == VIP_SUCCESS);	// Clear the completion queue.

	return retCode;
}


/*
 * cBundle::_HandleSend()
 *
 * Purpose:	Takes care of a packet pulled off of send queue.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	
 * Return:	
 */
void cBundle::_HandleSend(cConnection *connection, cPacket *packet)
{
	Header	   *header;

	header = packet->GetHeader();

	switch(header->opcode)
	{
		case _AM_ACK:
//			printf("RETURNING PACKET [ack-out]: %x\n", packet);
			connection->ReleasePacket(packet);
			break;
	}
}


/*
 * cBundle::_ConnectMain()
 *
 * Purpose:	The thread entry point for the bundle's connection thread.
 * IN:		-
 * OUT:		-
 * Cond:	A thread has been created w/ entry point ConnectMain.
 * PostCnd:	_ConnectMain() will loop until the thread is terminated.
 * Return:	Proper thread exit return code.
 */
DWORD cBundle::_ConnectMain()
{
	DWORD		 retCode = 0;
	VIP_RETURN	 nRet;
	cConnection	*connection;
	cConnection *extinctConnection;
	int			 numClosedConnections;

		// Get a new connection.
		connection = GetNewConnection();
		if(!connection)
		{
			AM_Error("Unable to get a free connection in _ConnectMain: Increase resources.", __LINE__, __FILE__);
			return retCode;
		}

		nRet = VIP_SUCCESS;
		while(mRunConnection)
		{
			nRet = connection->Listen(mLocalAddress, _LISTEN_TIMEOUT);
			if(nRet == VIP_SUCCESS)
			{
				// Get a new free connection.
				connection = GetNewConnection();
				if(!connection)
				{
					AM_Error("Unable to get a free connection in _ConnectMain: Increase resources.", __LINE__, __FILE__);
					break;
				}
			}

			// Check for extinct connections.
			numClosedConnections = mClosedConnectionQ->GetNumElements();
			for(int i = 0; i < numClosedConnections; i++)
			{
				if(mClosedConnectionQ->Pop((void **)&extinctConnection) != NULL)
				{
					if(GetTickCount() - extinctConnection->GetTimeStamp() < 1000)
					{
						mClosedConnectionQ->Push(extinctConnection);
					}
					else if(extinctConnection->IsExtinct())
					{
						printf("Deleting connection: %x <<EXTINCTION>>\n", extinctConnection);
						delete extinctConnection;				
					}
					else if(GetTickCount() - extinctConnection->GetTimeStamp() > 5000)
					{
						printf("Deleting connection: %x <<DUE TO EXTINCTION TIMEOUT>>\n", extinctConnection);
						delete extinctConnection;
					}
					else
					{
						mClosedConnectionQ->Push(extinctConnection);
					}
				}
			}
		} // end while(mRun)	
	return retCode;
}

/*
 * cBundle::ConnectThread()
 *
 * Purpose:	The thread entity that will call _ConnectMain.
 * IN:		A pointer to the cBundle.
 * OUT:		-
 * Cond:	A thread has been created w/ entry point ConnectThread.
 * PostCnd:	Will wait until return from _ConnectMain.
 * Return:	Proper thread exit return code.
 */
DWORD WINAPI ConnectThread(LPVOID lpParameter)
{
	DWORD		 retCode = 0;
	cBundle		*bundle;
//	printf("Inside of connect thread...whoopee!\n");
	
	//Start up the bundle connect
	bundle  = (cBundle *)lpParameter;
	retCode = bundle->_ConnectMain();

	//Shut down the thread.
	AM_Error("ConnectThread(): Shutting down thread.", __LINE__, __FILE__, AM_INFO);
	ExitThread(retCode);

	//To avoid warning
	return retCode;
}

/*
 * cBundle::BundleThread()
 *
 * Purpose:	The thread entity that will call _BundleMain.
 * IN:		A pointer to the cBundle::_BundleMain() function.
 * OUT:		-
 * Cond:	A thread has been created w/ entry point BundleThread.
 * PostCnd:	Will wait until return from _BundleMain.
 * Return:	Proper thread exit return code.
 */
DWORD WINAPI BundleThread(LPVOID lpParameter)
{
	DWORD		 retCode = 0;
	cBundle		*bundle;
//	printf("Inside of thread...whoopee!\n");
	
	//Start up the bundle
	bundle  = (cBundle *)lpParameter;
	retCode = bundle->_BundleMain();

	//Shut down the thread.
	AM_Error("BundleThread(): Shutting down thread.", __LINE__, __FILE__, AM_INFO);
	ExitThread(retCode);

	//To avoid warning
	return retCode;
}

/*
 * cBundle::GetNewConnection()
 *
 * Purpose:	Returns a new connection if possible.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	A new connection is created.
 * Return:	NULL if fail, else a cConnection pointer.
 */
cConnection* cBundle::GetNewConnection()
{
	cConnection *connection = NULL;
	Node		*node;

	try
	{
		connection = new cConnection(mAMLayer->GetNic(), mCQ, this, NULL);
	}
	catch(AMException& e)
	{
		AM_Error(e.ToString(), __LINE__, __FILE__);
	}
	if(connection)
	{
		node = mUsedConnectionQ->Push((void *)connection);
		if(node)
		{
			connection->mNode = node;
		}
		else
		{
			delete connection;
			connection = NULL;
		}
	}
	return connection;
}

int cBundle::GetNumber() { return mNumber; }