/*
 * cConnection.cpp
 * 
 * Class implementation for Connection.
 *
 * Authors:  Vicky & Ted
 * Last Mod: 4 March 1999 by ted
 */
#include "cConnection.h"
#include "cBundle.h"
#include "cEndpoint.h"
#include "cPacket.h"
#include "cHashTable.h"

#include "assert.h"

extern cHashTable* gAMConnectionTable;

/*
 * cConnection::cConnection()
 *
 * Purpose:	Create and initialize a new connection.
 * IN:		nic		-> Pointer to the nic
 *			cq		-> Pointer to the connection's completion queue.
 *			budne	-> The bundle tht this connection is in.
 * OUT:		-
 * Cond:	-
 * PostCnd:	A VI and a VIAddr are allocated for this connection.
 * Throws	AMException
 * Return:	-
 */
cConnection::cConnection(cNic *nic, cCQ *cq, cBundle *bundle, cEndpoint *endpoint, int outstandingReq)
{
	cPacket	*createPacket;

	mVI			= NULL;
	mVIAddress	= NULL;
	mSendWindow = NULL;
	mRecvWindow = NULL;
	mNode		= NULL;
	mBundle		= bundle;
	mEndpoint	= endpoint;
	mActive		= false;
	mEnabled	= false;
	mFreePacketQ= NULL;
	mNumPackets = 2*outstandingReq;

	mSendWindowInsert  = 0;
	mSendWindowRemove  = 0;
	mRecvWindowPtr     = 0;
	mRecvPacketsPosted = 0;

	mSeqOut		= 1;
	mCumSeqOut	= 0;
	mCumSeqIn	= 0;

	// Create the VI.
	mVI = _MakeVI(nic, cq);
	if(!mVI)
	{
		throw AMException("Unable to make VI in cConnection::cConnection.");
	}

	// Create the VIAddr.
	mVIAddress = nic->CreateVIAddr();
	if(!mVIAddress)
	{
		throw AMException("Unable to make VIAddr in cConnection::cConnection.");
	}
	
	// Create the send and receive windows.
	mRecvWindow = new cPacket*[outstandingReq];
	mSendWindow = new cPacket*[outstandingReq];
	if(!mRecvWindow || !mSendWindow)
	{
		throw AMException("Unable to create send/recv window in connection.");
	}
	for(unsigned int i = 0; i < (unsigned int)outstandingReq; i++)
	{
		mRecvWindow[i] = NULL;
		mSendWindow[i] = NULL;
	}
	mRecvWindowSize		 = outstandingReq;
	mSendWindowSize		 = outstandingReq;
	mSendWindowAvailable = outstandingReq;
	memset(&mAMAddress, 0, sizeof(_en_t));

	// Create the free packet queue and fill it.
	try 
	{
		mFreePacketQ = new cStack(mNumPackets);
		if(!mFreePacketQ)
		{
			throw AMException("Unable to create free packet queue in cConnection::cConnection.");
		}
	}
	catch(cStackException &e)
	{
		throw AMException(e.ToString());
	}
	for(i = 0; i < mNumPackets; i++)
	{
		createPacket = new cPacket(AM_MAX_MEDIUM, nic, bundle->GetPtag());
		if(!createPacket)
		{
			throw AMException("Unable to create free packet in cConnection::cConnection.");
		}
		else
		{
			ReleasePacket(createPacket);
		}
	}
}

/*
 * cConnection::_ResizeWindows()
 *
 * Purpose:	Attempts to dynamically resize the send/receive windows.
 * IN:		-
 * OUT:		-
 * Cond:	Only one process can call RecvPacket at a time!
 * PostCnd:	The windows are resized, if possible.
 * Return:	true if success, else fail.
 */
bool cConnection::_ResizeWindows(unsigned int newSize, int *currentPtr)
{
	cPacket **newSendWindow;
	cPacket **newRecvWindow;
	cPacket  *packet;
	int		  newNumPackets;

	// Not too concerned about overallocating resources.
	if(mSendWindowSize > newSize)
	{
		// Everything is already OK
		return true;
	}

//	printf("Resizing window to size = %d\n", newSize);

	// Attempt to create new arrays for send/recv windows.
	newSendWindow = new cPacket*[newSize];
	if(!newSendWindow) { return false; }
	newRecvWindow = new cPacket*[newSize];
	if(!newRecvWindow) { delete newSendWindow; return false; }
	for(unsigned int i = 0; i < newSize; i++)
	{
		newSendWindow[i] = NULL;
		newRecvWindow[i] = NULL;
	}

	// If the free packet queue is not big enough, create a new one.
	newNumPackets = 2*newSize;
	if(!mFreePacketQ->Resize(newNumPackets))
	{
		delete newSendWindow;
		delete newRecvWindow;
		return false;
	}
	for(i = 0; i < (newNumPackets-mNumPackets); i++)
	{
		packet = new cPacket(AM_MAX_MEDIUM, mBundle->GetNic(), mBundle->GetPtag());
		if(!packet)
		{
			return false;
		}
		ReleasePacket(packet);	// Push onto free queue.
	}

	// Copy over receive window.
	int tempRecvPtr = *currentPtr;
	int tempNewRecvPtr = 0;
	for(i = 0; i < mRecvWindowSize; i++)
	{
		newRecvWindow[tempNewRecvPtr] = mRecvWindow[tempRecvPtr];
		tempNewRecvPtr++;
		tempRecvPtr = (tempRecvPtr + 1) % mRecvWindowSize;
		*currentPtr = 0;  // The first entry is the one being processed -> This will automatically
						  // set the mRecvWindowPtr for us.
	}
	mRecvWindowSize = newSize;
	delete[] mRecvWindow;	// Free up the mem.
	mRecvWindow = newRecvWindow;

	// Update the new send window.
	mSendCS.Enter();

	// Post appropriate number of recvs (avoid race condition where this connection
	//									 is being killed at this very moment!)
	if(mActive)
	{
		for(i = 0; i < (newNumPackets-mNumPackets)/2; i++)
		{
			PostRecv();
		}
	}

	int j = 0;
	while(mSendWindowRemove != mSendWindowInsert)
	{
		newSendWindow[j] = mSendWindow[mSendWindowRemove];
		mSendWindowRemove = (mSendWindowRemove + 1) % mSendWindowSize;
		j++;
	}
	mSendWindowAvailable = newSize - (mSendWindowSize - mSendWindowAvailable);
	delete[] mSendWindow;	// Free up the mem.
	mSendWindow = newSendWindow;
	mSendWindowRemove = 0;
	mSendWindowInsert = j;
	mSendWindowSize   = newSize;

	mSendCS.Leave();

	mNumPackets = newNumPackets;

	return true;
}

/*
 * cConnection::~cConnection()
 *
 * Purpose:	Destroys the connection.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The connection is unuseable.
 * Return:	-
 */
cConnection::~cConnection()
{
	mSendCS.Enter();
	mRecvCS.Enter();

	cPacket	*packet;

	if(mVI)			{ delete mVI; mVI = NULL; }
	if(mVIAddress)	{ delete mVIAddress;  mVIAddress  = NULL; }
	if(mSendWindow)	{ delete mSendWindow; mSendWindow = NULL; }
	if(mRecvWindow)	{ delete mRecvWindow; mRecvWindow = NULL; }

	// Clean up the free packet queue.
	if(mFreePacketQ)
	{
		while(mFreePacketQ->Pop((void **)&packet))
		{
			delete packet;
		}
		delete mFreePacketQ;
		mFreePacketQ = NULL;
	}

	mRecvCS.Leave();
	mSendCS.Leave();
}

/*
 * cConnection::RecvPacket()
 *
 * Purpose:	Receives a packet.
 * IN:		packet		-> The packet that is received.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The packet is placed in the receive window.  An ack is sent.  Packet may be processed.
 * Return:	AM condition code.
 */
int cConnection::RecvPacket(cPacket *packet)
{
	unsigned int  seqNum;
	Header		 *header;
	cEndpoint	 *tempEP;

	header = packet->GetHeader();

//	printf("Received packet.  (SeqNum=%d, opcode=%d)\n", header->sequenceNum, header->opcode);

	mRecvPacketsPosted--;

	// CAUTION: MAY NEED TO CHANGE BELOW!!
//	if(header->opcode == _AM_REQUEST)
//	{
		PostRecv();	//Post a recv so we don't run out.
//	}

	switch(header->opcode)
	{
		case _AM_ACK:
			_HandleAck(header);
			ReleasePacket(packet);
			break;

		case _AM_REPLY:
		case _AM_CONNECT_REPLY:
			_HandleAck(header);
			// FALL THRU
		case _AM_REQUEST:
		case _AM_CONNECT:
		case _AM_XFER_INTERMEDIATE:
			seqNum = header->sequenceNum;
			if(seqNum <= mCumSeqIn)
			{
				ReleasePacket(packet);  // Already seen packet, toss it.
			}
			else
			{
				unsigned int ptr;
				unsigned int difference = (seqNum - (mCumSeqIn + 1));
				assert(difference < (mRecvWindowSize+1));
	
				ptr = (mRecvWindowPtr + difference) % mRecvWindowSize;
				mRecvWindow[ptr] = packet;
				_AdvanceReceiveWindow();
			}
			break;

		case _AM_RTS:
			_HandleAck(header);
			tempEP = mBundle->GetEndpoint(header->srcEndpoint);
			printf("Received RTS...checking ep.\n");
			if(tempEP)
			{
				printf("Calling processRTS.\n");
				tempEP->ProcessRTS(packet);
			}
			ReleasePacket(packet);
			break;
	} /* switch */

	return 1;
}

/*
 * cConnection::AdvanceReceiveWindow()
 *
 * Purpose:	Advances the receive window pointer if possible.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	Some packets may be processed.
 * Return:	AM condition code.
 */
int	cConnection::_AdvanceReceiveWindow()
{
	int			 ptr;
	cPacket		*packet;
	Header		*header;
	bool		 sendAck = true;

	ptr = mRecvWindowPtr;
	while(mRecvWindow[ptr] != NULL)
	{
		mCumSeqIn++;
		packet = mRecvWindow[ptr];
		header = packet->GetHeader();

		// Set endpoint for packet if necessary.
		if(!mEndpoint)
		{
			mEndpoint = mBundle->GetEndpoint(header->destEndpoint);
			if(!mEndpoint)
			{
				// RETURN TO SENDER!!!
			}
		}

		// Handle packet.
		switch(header->opcode)
		{
			case _AM_REQUEST:
				if(mEndpoint->ProcessRequest(packet) != AM_OK)
				{
					header->opcode = _AM_RTS;
					packet->Reset();
					packet->SetLength(header->nSize);
					SendPacket(packet);
				}
				else
				{
					ReleasePacket(packet);
				}
				sendAck = false;
				break;
			case _AM_REPLY:
				if(mEndpoint->ProcessReply(packet) != AM_OK)
				{
					header->opcode = _AM_RTS;
					SendPacket(packet);
					sendAck = false;
				}
				else
				{
					ReleasePacket(packet);
				}
				sendAck = true;
				break;

			case _AM_XFER_INTERMEDIATE:
				mEndpoint->ProcessXferIntermediate(packet);
				ReleasePacket(packet);
				sendAck = true;
				break;

			case _AM_CONNECT:
//				printf("Received _AM_CONNECT!\n");
				_FinishConnection(packet, &ptr);
				sendAck = false;
				// No release packet since we re-use for send back.
				break;

			case _AM_CONNECT_REPLY:
//				printf("Received _AM_CONNECT_REPLY!\n");
				mConnectStatus = header->args[1];
//				printf("mConnectStatus = %d\n", mConnectStatus);
				if(mConnectStatus != _CONNECT_STATE_SUCCESS)
				{
					mEnabled = false;	// Disable the connection.
				}
				ReleasePacket(packet);
				sendAck = true;
				break;
		}
		mRecvWindow[ptr] = NULL;
		ptr = (ptr + 1) % mRecvWindowSize;
	}
	mRecvWindowPtr = ptr;

	// SEND ACKNOWLEDGEMENT w/ new cumulative!
	if(sendAck)
	{
		_SendAck();
	}

	return 1;
}


/*
 * cConnection::SendPacket()
 *
 * Purpose:	Queues a packet for sending.
 * IN:		packet	-> The packet to send.
 * OUT:		-
 * Cond:	-
 * PostCnd:	1 extra packet is posted on recv queue for possible ack.  Packet is queued.
 * Return:	AM_OK || AM_ERR_RESOURCE <connLost> || AM_ERR_NOT_SENT <no room in send q> 
 */
int cConnection::SendPacket(cPacket *packet)
{
	Header		*header;
	VIP_RETURN	 rc;

	// Make sure the connection is up and running.
	if(!mEnabled)
	{
		AM_Error("cConnection::SendPacket -> Attempt to send on non-active connection.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		ReleasePacket(packet);
		return AM_ERR_RESOURCE;
	}

	if(mSendWindowAvailable == 0)
	{
		AM_Error("cConnection::SendPacket -> Rejected due to no window room.", __LINE__, __FILE__, AM_ERROR);
		ReleasePacket(packet);
		return AM_ERR_NOT_SENT;
	}

	// Set the packet sequence number
	header = packet->GetHeader();
	header->hidden.connPtr = this;
	header->cumulativeAck = mCumSeqIn;	// Basically piggy back an ack

	mSendCS.Enter();		// LOCK: Send window

		header->sequenceNum = mSeqOut;
		mSeqOut++;

		// Post the packet to the connection.
		rc = mVI->PostSend(packet->GetDescPtr(), packet->GetDescMemHandle());
		if(rc != VIP_SUCCESS)
		{
			mSeqOut--;
			mSendCS.Leave();
			ReleasePacket(packet);
			return AM_ERR_RESOURCE;
		}

		// Insert the packet into the send window.
		mSendWindow[mSendWindowInsert] = packet;
		mSendWindowInsert = (mSendWindowInsert + 1) % mSendWindowSize;
		mSendWindowAvailable--;

	mSendCS.Leave();	// UNLOCK: Send window
	return AM_OK;
}

/*
 * cConnection::_HandleAck
 *
 * Purpose:	Handles acknowledgements.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	Packets may be released from the send queue.
 * Return:	AM condition code.
 */
int	cConnection::_HandleAck(Header *header)
{
	unsigned int	ackCumulative;
	cPacket		   *packet;
	
	ackCumulative = header->cumulativeAck;

//	printf("Received ACK (ackCum=%d)\n", ackCumulative);
//	fflush(stdout);

	if(ackCumulative < mCumSeqOut)
	{
		return 1;	// Drop ack.
	}
	else
	{
		mSendCS.Enter();		// LOCK: Send window

		if(ackCumulative > mCumSeqOut)
		{
			int difference = (ackCumulative - mCumSeqOut);
			for(int i = 0; i < difference; i++)
			{
				mCumSeqOut++;
				packet = mSendWindow[mSendWindowRemove];
				mSendWindow[mSendWindowRemove] = NULL;

				// Free the packet.
				ReleasePacket(packet);

				mSendWindowRemove = (mSendWindowRemove + 1) % mSendWindowSize;
				mSendWindowAvailable++;
			}
		}
		else
		{
			// PROBABLY DROPPED A PACKET -> Fast Retransmit!!
			printf("<<<NEED TO RETRANSMIT HERE!!>>>\n");
			// i.e. start at remove pointer, then 
		}

		mSendCS.Leave();	// UNLOCK: Send window
	}
	
	return 1;
}

/*
 * cConnection::_SendAck()
 *
 * Purpose:	Sends an acknowledgement.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true if success, else false
 */
int	cConnection::_SendAck()
{
//	printf("Sending Ack (mCumSeqIn=%d)\n", mCumSeqIn);

	cPacket *packet;
	Header	*header;
	VIP_RETURN	rc;

	packet = GetFreePacket();
	if(!packet)
	{
		AM_Error("_SendAck(): Unable to get free packet for ack!", __LINE__, __FILE__, AM_ERROR);
		return AM_ERR_RESOURCE;
	}
	header = packet->GetHeader();
	header->opcode = _AM_ACK;
	header->cumulativeAck = mCumSeqIn;

	// Make it so that the packet can be returned after send
	header->hidden.connPtr   = this;

	// Set up the packet.
	packet->SetLength(0);

	// Post the ack.
	rc = mVI->PostSend(packet->GetDescPtr(), packet->GetDescMemHandle());
	if(rc != VIP_SUCCESS)
	{
		return AM_ERR_NOT_SENT;
	}
	else 
	{
		return AM_OK;
	}
}

/*
 * cConnection::OpenConnection()
 *
 * Purpose:	Attempts to establish a connection with a remote VI.
 * IN:		remoteAddress	-> The remote address to connect to.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The connection becomes valid.
 * Return:	true if success, else false
 */
bool cConnection::OpenConnection(en_t remoteAddress)
{
	VIP_RETURN	rc;
	cHashTable *connTable;
	cPacket	   *packet;
	Header	   *header;
	int			time;

	connTable = gAMConnectionTable;
	if(!connTable->Insert(mVI->GetHandle(), this))
	{
		AM_Error("Ran out of space in the connection table!", __LINE__, __FILE__);
		return false;
	}
	for(unsigned int i = 0; i < mSendWindowSize; i++)
	{
		if(!PostRecv())
		{
			AM_Error("Unable to post initial recvs in OpenConnection.", __LINE__, __FILE__);		
		}
	}

	// Set up the address.
	mVIAddress->GetHostByName(remoteAddress->node);
	mVIAddress->SetDiscriminator(remoteAddress->unique, sizeof(AM_NAME_UNIQUE_SIZE+AM_NAME_BUNDLE_SIZE));

	// Make connection.
	for(time = 0; time < 20000; time+=2000)
	{
		rc = mVI->ConnectRequest(mBundle->mLocalAddress, mVIAddress, 2000);
		if (rc == VIP_TIMEOUT)
		{
			continue;
		}
		else if (rc != VIP_SUCCESS)
		{
			return false;
		}
		else
		{
			printf("{OPENED CONNECTION!}\n");
			memcpy(&mAMAddress, remoteAddress, sizeof(_en_t));
			mActive = true;
			mEnabled = true;

			// Pass connection parameters (optimistic!)
			mConnectStatus = _CONNECT_STATE_WAITING;
			packet = GetFreePacket();
			if(!packet) { return false; }
			header = packet->GetHeader();
			header->opcode  = _AM_CONNECT;
			header->nSize   = 0;
			packet->SetLength(0);
			header->args[0] = mEndpoint->GetExpectedMaxNumRequests();
			SendPacket(packet);
			return true;
		}
	} /* end for */
	return false;	// Must have run out of time.
}

/*
 * cConnection::_FinishConnection()
 *
 * Purpose:	Attempts to allocate enough resources for a connection (called on recv connection side!)
 * IN:		packet		-> the _AM_CONNECT packet.
 * OUT:		-
 * Cond:	-
 * PostCnd:	If success, the packet is ack'd.  If fail it is nack'd.
 * Return:	VIP_RETURN
 */
int	cConnection::_FinishConnection(cPacket *packet, int *currentPtr)
{
	Header *header;
	int		outstandingReq;
	int		reply;

	header = packet->GetHeader();
	outstandingReq = header->args[0];

	if(_ResizeWindows(outstandingReq, currentPtr))
	{
		reply = _CONNECT_STATE_SUCCESS;
	}
	else
	{
		reply = _CONNECT_STATE_FAIL;
	}

	// Generate reply
	packet->Reset();
	header->opcode  = _AM_CONNECT_REPLY;
	header->nSize   = 0;
	packet->SetLength(0);
	header->args[1] = reply;
	SendPacket(packet);

	return 1;
}

/*
 * cConnection::Listen()
 *
 * Purpose:	Listens for a new connection on this cConnection.
 * IN:		localAddress	-> The local VI address (for setting discriminator)
 *			timeout			-> The time to wait for a new connection.
 * OUT:		-
 * Cond:	-
 * PostCnd:	If success, the connection is 
 * Return:	VIP_RETURN
 */
VIP_RETURN cConnection::Listen(cVIAddr *localAddress, long timeout)
{
	VIP_RETURN	rc;
	cHashTable *connTable;

	// Post necessary recv packets in advance.
	for(unsigned int i = mRecvPacketsPosted; i < mRecvWindowSize; i++)
	{
		PostRecv();
	}

	rc = mVI->ConnectWait(localAddress, mVIAddress, timeout);
	if(rc == VIP_SUCCESS)
	{

		// Insert this connection into the connection table.
		connTable = gAMConnectionTable;
		if(!connTable->Insert(mVI->GetHandle(), this))
		{
			AM_Error("Ran out of space in the connection table!", __LINE__, __FILE__);
		}

		mEnabled = true;
		mActive	 = true;

		rc = mVI->ConnectAccept();
		if(rc != VIP_SUCCESS)
		{
			connTable->Delete(mVI->GetHandle());
			mEnabled = false;
			mActive  = false;
			return rc;
		}
		
		// Set up mAMAddress!
		unsigned long length = AM_MAX_NODE_NAME_LENGTH;
		mVIAddress->GetHostByAddr((char *)&(mAMAddress.node), &length);
		int disclength = AM_NAME_UNIQUE_SIZE+AM_NAME_BUNDLE_SIZE;
		mVIAddress->GetDiscriminator((char *)&(mAMAddress.unique), &disclength);

//		printf("{CONN %x RECEIVED CONNECTION}: %s:%s:%s\n", this, mAMAddress.node, mAMAddress.unique, mAMAddress.bundle);
	}
	return rc;
}

/*
 * cConnection::CloseConnection()
 *
 * Purpose:	Gracefully closes the connection with the remote end.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The connection is closed and available for re-use.
 * Return:	true if success, else false
 */
bool cConnection::CloseConnection()
{
	Header			*header;
	cPacket			*packet;
	cHashTable		*connTable;
	unsigned int	 i;
//	DWORD			 startTick;

	mSendCS.Enter();

	if(!mActive)
	{
		AM_Error("Attempt to close connection 2x!", __LINE__, __FILE__);
		mSendCS.Leave();
		return true;
	}

	// Make this connection no longer active!
	mActive  = false;
	mEnabled = false;

	// Disconnect the VI
	mVI->Disconnect();

	mSendCS.Leave();

	printf("Closing connection %x\n", this);

	// Delete all packets left in the receive window.
	for(i = 0; i < mRecvWindowSize; i++)
	{
		packet = mRecvWindow[i];
		if(packet)
		{
			ReleasePacket(packet);
		}
	}

	// Return all packets left in the send window to the sender.
	while(mSendWindowAvailable < mSendWindowSize)
	{
		packet = mSendWindow[mSendWindowRemove];
		assert(packet);
		header = packet->GetHeader();

		// Return to sender.
		mEndpoint = mBundle->GetEndpoint(header->srcEndpoint);
		if(mEndpoint)
		{
			printf("RTS: Due to packet pulled out of send queue.\n");
			header->status = EUNREACHABLE;
			mEndpoint->ProcessRTS(packet);
		}

		// Free the packet.
		ReleasePacket(packet);

		mSendWindowRemove = (mSendWindowRemove + 1) % mSendWindowSize;
		mSendWindowAvailable++;
	}

	// Remove from connection table.
	connTable = gAMConnectionTable;
	if(!connTable->Delete(mVI->GetHandle()))
	{
		AM_Error("Unable to remove connection from connection table!", __LINE__, __FILE__);
	}

	// Remove from connection queue if necessary.
	if(mNode)
	{
		printf("Removing connection from connFifo.\n");
		cFifo *connFifo = mBundle->GetConnectionQ();
		connFifo->Delete(mNode);
	}

	// Add to pending extinct connection queue.
	mTimeStamp = GetTickCount();
	mBundle->AddToClosedConnectionQueue(this);

//	printf("Connection closed.\n");

	return true;
}

/*
 * cConnection::PostRecv()
 *
 * Purpose:	Adds another packet to the VI's recv queue.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	Another packet is waiting on the recv queue.
 * Return:	true if success, else false
 */
bool cConnection::PostRecv()
{
	Header		*header;
	cPacket		*packet;
	VIP_RETURN	 rc;

//	printf("><Posting Recv><\n");

	// Post room for an ack.
	packet = GetFreePacket();
	if(!packet)
	{
		AM_Error("cConnection::PostRecv-> Unable to get free packet.", __LINE__, __FILE__, AM_ERROR);
		return false;
	}
	header = packet->GetHeader();
	header->hidden.connPtr = this;
	rc = mVI->PostRecv(packet->GetDescPtr(), packet->GetDescMemHandle());
	if(rc != VIP_SUCCESS)
	{
		ReleasePacket(packet);
		return false;
	}
	mRecvPacketsPosted++;
	return true;
}

/*
 * cConnection::IsExtinct()
 *
 * Purpose:	Returns true if this connection's queues are empty.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	true if queues are empty, else false.
 */
bool cConnection::IsExtinct()
{
	return (mVI->IsSendQEmpty() && mVI->IsRecvQEmpty());
}

/*
 * cConnection::SetAMAddress()
 *
 * Purpose:	Sets the am address that this is connected to.
 * IN:		addr	-> The AM address to set this to.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The internal remote AM address is set.
 * Return:	-
 */
void cConnection::SetAMAddress(en_t addr)
{
	memcpy(&mAMAddress, addr, sizeof(_en_t));
}

/*
 * cConnection::ReleasePacket()
 *
 * Purpose:	Returns a free packet to the free queue.
 * IN:		packet		-> The packet to return to the free queue.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An additional packet is on the free packet queue.
 * Return:	-
 */
void cConnection::ReleasePacket(cPacket *packet)
{
	packet->Reset();
	if(!mFreePacketQ->Push((void *)packet))
	{
		AM_Error("Unable to push free packet on stack.", __LINE__, __FILE__);
		delete packet;	// Make sure it gets cleaned up.
	}
}

/*
 * cConnection::_MakeVI()
 *
 * Purpose:	Creates a fresh new VI.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	NULL if fail, else a cVI pointer.
 */
cVI* cConnection::_MakeVI(cNic *nic, cCQ *cq)
{
	VIP_VI_ATTRIBUTES		  viAttrib;
	const VIP_NIC_ATTRIBUTES *nicAttrib;

	nicAttrib = nic->GetAttrib();

	viAttrib.ReliabilityLevel = VIP_SERVICE_UNRELIABLE;
	viAttrib.MaxTransferSize  = nicAttrib->MaxTransferSize;
	viAttrib.QoS			  = 0;
	viAttrib.Ptag			  = mBundle->GetPtag()->GetPtag();
	viAttrib.EnableRdmaWrite  = VIP_FALSE;
	viAttrib.EnableRdmaRead   = VIP_FALSE;

	return(nic->CreateVI(&viAttrib, cq, cq));
}