/*
 * cEndpoint.cpp
 * 
 * Class implementatino for AM endpoint class. 
 *
 * Authors:  Vicky & Ted
 */
#include "cEndpoint.h"

#include "cPacket.h"
#include "cBundle.h"
#include "cConnection.h"

//Std handler functions
static void abort();

/*
 * cEndpoint::cEndpoint()
 *
 * Purpose:	Create and initialize a new endpoint.
 * IN:		bundle	-> Pointer to the bundle this endpoint is currently in.
 * OUT:		-
 * Cond:	The given bundle exists, etc.
 * PostCnd:	The endpoint's data structures are alloc'd and initialized.
 * Throws:	AMException if creation fails.	
 * Return:	-
 */
cEndpoint::cEndpoint(cBundle *bundle, int number)
{
	mBundle  = bundle;
	mNumber  = number;
	mMapSize = DEFAULT_ENDPOINTNAME_ARRAY_SIZE;
	mHandlerTableSize = DEFAULT_HANDLERTABLE_ARRAY_SIZE;
	mNumOutstandingRequests = DEFAULT_WINDOW_SIZE;

	mMap			= NULL;
	mHandlerTable	= NULL;
	mTag			= AM_NONE;

	// Set up the map table.
	mMap = new MapEntry[mMapSize];
	if(!mMap)
	{
		throw AMException("<cEndpoint::cEndpoint> Unable to alloc space for map entries.");
	}
	for(unsigned int i=0; i<mMapSize; i++)
	{
		mMap[i].tag  = AM_NONE;		//Init all tages to none.
		mMap[i].connection = NULL;
	}

	// Set up the handler table.
	mHandlerTable = new HandlerEntry[mHandlerTableSize];
	if(!mHandlerTable)
	{
		throw AMException("<cEndpoint::cEndpoint> Unable to alloc space for handler entries.");
	}
	for(i=0; i<mHandlerTableSize; i++)
	{
		mHandlerTable[i].function = abort;
	}
}

/*
 * cEndpoint::~cEndpoint()
 *
 * Purpose:	Destroy the endpoint.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	The endpoints resources are de-allocated and destroyed.
 * Return:	-
 */
cEndpoint::~cEndpoint()
{
	mCSMap.Enter();
	// Delete anything in the map table.
	if(mMap)
	{
		for(unsigned int i=0; i<mMapSize; i++)
		{
			if(mMap[i].connection != NULL)
			{
				Unmap(i);
			}
		}
		delete[] mMap;
	}

	// Delete anything in the handler table.
	if(mHandlerTable)
	{
		delete[] mHandlerTable;
	}
	mCSMap.Leave();
}

/*
 * cEndpoint::SetExpectedResources()
 *
 * Purpose:	Attempts to set the expected resources value.
 * IN:		n_endpoints				-> Number of endpoints that this can support.
 *			n_outstanding_requests	-> Number of outstanding requests supported at a time.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The maptable may be increased in size if necessary.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::SetExpectedResources(int n_endpoints, int n_outstanding_requests)
{
	if((unsigned int)n_endpoints > mMapSize)
	{
		return AM_ERR_RESOURCE;
	}
	if(n_outstanding_requests < 1)
	{
		return AM_ERR_BAD_ARG;
	}
	mNumOutstandingRequests = n_outstanding_requests;
	return AM_OK;
}

/*
 * cEndpoint::Map()
 *
 * Purpose:	Map a remote address into the endpoints MapTable.
 * IN:		epn		-> The remote address to map.
 *			tag		-> The tag to associate with the remote endpoint.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The remote endpoint is mapped into the index.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int	cEndpoint::Map(int index, en_t epn, tag_t tag)
{
	cConnection *newConnection = NULL;

	// Sanity
	if((unsigned int)index > mMapSize)
	{
		return AM_ERR_BAD_ARG;
	}

	// Create a new connection object.
	try
	{
		newConnection = new cConnection(mBundle->GetNic(), mBundle->GetCQ(), mBundle, this, mNumOutstandingRequests);
	}
	catch(AMException &e)
	{
		AM_Error(e.ToString(), __LINE__, __FILE__);
		if(newConnection) { delete newConnection; newConnection = NULL; }
	}
	if(!newConnection)
	{
		return AM_ERR_RESOURCE;
	}

	// Attempt to connect.
	if(!newConnection->OpenConnection(epn))
	{
		delete newConnection;
		return AM_ERR_RESOURCE;
	}

	// Insert entry
	mCSMap.Enter();

	mMap[index].tag			= tag;
	mMap[index].number		= epn->endpoint;
	mMap[index].connection	= newConnection;

	mCSMap.Leave();

	return AM_OK;
}

/*
 * cEndpoint::MapAny()
 *
 * Purpose:	Map a remote address into the endpoints MapTable.
 * IN:		epn		-> The remote address to map.
 *			tag		-> The tag to associate with the remote endpoint.
 * OUT:		index	-> The index at which the endpoint was mapped.
 * Cond:	-
 * PostCnd:	The remote endpoint is mapped into the index.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int	cEndpoint::MapAny(int *index, en_t epn, tag_t tag)
{
	cConnection *newConnection = NULL;
	int			 retVal = AM_ERR_RESOURCE;

	// Find entry.
	mCSMap.Enter();

	for(unsigned int i = 0; i < mMapSize; i++)
	{
		if(mMap[i].connection == NULL)
		{
			*index = i;
			retVal = Map(i, epn, tag);
			break;
		}
	}

	mCSMap.Leave();
	return retVal;
}

/*
 * cEndpoint::Unmap()
 *
 * Purpose:	Unmaps a remote address from the MapTable.
 * IN:		index	-> The table entry to unmap.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The MapTable is modified.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int	cEndpoint::Unmap(int index)
{
	if ((unsigned int)index > mMapSize)
	{
		return AM_ERR_BAD_ARG;
	}

	mCSMap.Enter();
	
	mMap[index].tag = AM_NONE;
	if(mMap[index].connection != NULL)
	{
		mMap[index].connection->CloseConnection();
//		delete mMap[index].connection;
		mMap[index].connection = NULL;
	}

	mCSMap.Leave();
	return AM_OK;
}

/*
 * cEndpoint::SetHandler()
 *
 * Purpose:	Sets the handler for the given index.
 * IN:		handler		-> The index into the handler table.
 *			function	-> The function to map as the handler.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The HandlerTable is modified.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::SetHandler(handler_t handler, void  (*function)())
{
	// Sanity
	if(handler > mHandlerTableSize)
	{
		return AM_ERR_BAD_ARG;
	}

	mCSHandler.Enter();
		mHandlerTable[handler].function = function;
	mCSHandler.Leave();
	return AM_OK;
}

/*
 * cEndpoint::SetHandlerAny()
 *
 * Purpose:	Sets the handler for the given index.
 * IN:		function	-> The function to map as the handler.
 * OUT:		handler		-> The index into the handler table.
 * Cond:	-
 * PostCnd:	The HandlerTable is modified.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::SetHandlerAny(handler_t *handler, void (*function)())
{
	mCSHandler.Enter();
	
	for(unsigned int i = 1; i < mHandlerTableSize; i++)
	{
		if(mHandlerTable[i].function == abort)
		{
			*handler = i;
			mHandlerTable[i].function = function;
			mCSHandler.Leave();
			return AM_OK;
		}
	}
	mCSHandler.Leave();
	return AM_ERR_RESOURCE;
}


/*
 * cEndpoint::SetNumHandlers()
 *
 * Purpose:	Sets the handler table size if possible.
 * IN:		newSize		-> The new size of the handler table.
 * OUT:		-
 * Cond:	-
 * PostCnd:	The HandlerTable is modified.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::SetNumHandlers(unsigned int newSize)
{
	HandlerEntry	*newHandlerTable;
	unsigned int	 i;

	mCSHandler.Enter();

	// If the new size is smaller, then see if we can manage.
	if(newSize < mHandlerTableSize)
	{
		// See if this would work
		for(i = newSize-1; i < mHandlerTableSize; i++)
		{
			if(mHandlerTable[i].function != abort)
			{
				mCSHandler.Leave();
				return AM_ERR_RESOURCE;
			}
		}
	}
	newHandlerTable = new HandlerEntry[newSize];
	if(!newHandlerTable)
	{
		mCSHandler.Leave();
		return AM_ERR_RESOURCE;
	}
	for(i = 0; i < mHandlerTableSize; i++)
	{
		newHandlerTable[i] = mHandlerTable[i];
	}
	for(i = mHandlerTableSize; i < newSize; i++)
	{
		newHandlerTable[i].function = abort;
	}
	delete[] mHandlerTable;
	mHandlerTable = newHandlerTable;
	mHandlerTableSize = newSize;

	mCSHandler.Leave();
	return AM_OK;
}

/*
 * cEndpoint::Reply4()
 *
 * Purpose:	Does a 4 arg AM reply.
 * IN:		token			-> The token that was given upon request.
 *			handler			-> The index into the remote handler table.
 *			a0..a3			-> The free args.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An AM is posted to the bundle.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::Reply4(void *token, handler_t handler, int a0, int a1, int a2, int a3)
{
	cConnection *connection;
	cPacket	*packet;
	Header	*header;
	Token	*ptoken;

	ptoken = (Token *)token;
	if(!ptoken)
	{
		return AM_ERR_BAD_ARG;
	}
	connection = ptoken->connection;

	// Get a free packet.
	packet = connection->GetFreePacket();
	if(!packet)
	{
		AM_Error("Unable to get free packet for reply.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		return AM_ERR_RESOURCE;
	}

	// Set up the packet.
	packet->SetLength(0);

	// Set up the header.
	header = packet->GetHeader();
	header->opcode = _AM_REPLY;
	header->handlerType = AM_REPLY_4;
	header->tag = ptoken->tag;
	header->args[0] = a0;
	header->args[1] = a1;
	header->args[2] = a2;
	header->args[3] = a3;
	header->destEndpoint = ptoken->source;
	header->srcEndpoint  = mNumber;
	header->handlerNum = handler;
	header->nSize = 0;
	return connection->SendPacket(packet);
}

/*
 * cEndpoint::ReplyI4()
 *
 * Purpose:	Does a 4 arg + buffer AM reply.
 * IN:		token			-> The token that was given upon request.
 *			handler			-> The index into the remote handler table.
 *			source			-> Buffer to send
 *			nbytes			-> Number of bytes to send from buffer.
 *			a0..a3			-> The free args.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An AM is posted to the bundle.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::ReplyI4(void *token, handler_t handler, void *source, int nbytes, int a0, int a1, int a2, int a3)
{
	cConnection *connection;
	cPacket	*packet;
	Header	*header;
	Token	*ptoken;

	ptoken = (Token *)token;
	if(!ptoken)
	{
		return AM_ERR_BAD_ARG;
	}
	connection = ptoken->connection;

	// Get a free packet.
	packet = connection->GetFreePacket();
	if(!packet)
	{
		AM_Error("Unable to get free packet for reply.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		return AM_ERR_RESOURCE;
	}

	// Set up the packet.
	packet->SetData(source, nbytes);

	// Set up the header.
	header = packet->GetHeader();
	header->opcode = _AM_REPLY;
	header->handlerType = AM_REPLY_I4;
	header->tag = ptoken->tag;
	header->args[0] = a0;
	header->args[1] = a1;
	header->args[2] = a2;
	header->args[3] = a3;
	header->destEndpoint = ptoken->source;
	header->srcEndpoint  = mNumber;
	header->handlerNum = handler;
	header->nSize = nbytes;
	return connection->SendPacket(packet);
}

/*
 * cEndpoint::ReplyXfer4()
 *
 * Purpose:	Does a 4 arg + buffer to remote VM segment AM reply.
 * IN:		token			-> The token that was given upon request.
 *			dest_offset		-> Offset in remote ep's vm segment to place the data.
 *			handler			-> The index into the remote handler table.
 *			source			-> Buffer to send
 *			nbytes			-> Number of bytes to send from buffer.
 *			a0..a3			-> The free args.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An AM is posted to the bundle.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::ReplyXfer4(void *token, int dest_offset, handler_t handler, void *source, int nbytes, int a0, int a1, int a2, int a3)
{
	cConnection *connection;
	cPacket	*packet;
	Header	*header;
	Token	*ptoken;
	int		 bytesRemaining = nbytes;
	int		 offset		    = 0;

	ptoken = (Token *)token;
	if(!ptoken)
	{
		return AM_ERR_BAD_ARG;
	}
	connection = ptoken->connection;

	// Fragment into AM_MAX_MEDIUM size chunks.
	while(bytesRemaining > 0)
	{
		// Get a free packet.
		packet = connection->GetFreePacket();
		if(!packet)
		{
			AM_Error("Unable to get free packet for reply.", __LINE__, __FILE__, AM_ERR_RESOURCE);
			return AM_ERR_RESOURCE;
		}

		// Get packet header.
		header = packet->GetHeader();

		// Set up part of header.
		if(bytesRemaining > AM_MAX_MEDIUM)
		{
			header->opcode = _AM_XFER_INTERMEDIATE;
			packet->SetData((char *)source+offset, AM_MAX_MEDIUM);
			header->nSize       = AM_MAX_MEDIUM;
			header->destOffset = dest_offset+offset;
		}
		else
		{
			header->opcode = _AM_REPLY;
			packet->SetData((char *)source+offset, bytesRemaining);
			header->handlerType = AM_REPLY_XFER4;
			header->nSize		= nbytes;
			header->destOffset	= dest_offset;
		}

		// Set up the header.
		header->tag = ptoken->tag;
		header->args[0] = a0;
		header->args[1] = a1;
		header->args[2] = a2;
		header->args[3] = a3;
		header->destEndpoint = ptoken->source;
		header->srcEndpoint  = mNumber;
		header->handlerNum = handler;
		if(connection->SendPacket(packet) != AM_OK)
		{
			printf("Error, unable to send XFER!\n");
		}

		offset+=header->nSize;
		bytesRemaining -= header->nSize;
	}
	return AM_OK;
}

/*
 * cEndpoint::Request4()
 *
 * Purpose:	Does a 4 arg AM request.
 * IN:		reply_endpoint	-> endpoint index to send to.
 *			handler			-> The index into the remote handler table.
 *			a0..a3			-> The free args.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An AM is posted to the bundle.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::Request4(int reply_endpoint, handler_t handler, int a0, int a1, int a2, int a3)
{
	cConnection *connection;
	cPacket	*packet;
	Header	*header;

	connection = mMap[reply_endpoint].connection;
	if(!connection)
	{
		return AM_ERR_BAD_ARG;
	}

	// Get a free packet.
	packet = connection->GetFreePacket();
	if(!packet)
	{
		AM_Error("Unable to get free packet for send.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		return AM_ERR_RESOURCE;
	}

	// Set up the packet.
	packet->SetLength(0);

	// Set up the header.
	header = packet->GetHeader();
	header->opcode = _AM_REQUEST;
	header->handlerType = AM_REQUEST_4;
	header->tag = mMap[reply_endpoint].tag;
	header->args[0] = a0;
	header->args[1] = a1;
	header->args[2] = a2;
	header->args[3] = a3;
	header->destEndpoint = mMap[reply_endpoint].number;
	header->srcEndpoint  = mNumber;
	header->handlerNum = handler;
	header->nSize = 0;
	return connection->SendPacket(packet);
}

/*
 * cEndpoint::RequestI4()
 *
 * Purpose:	Does a 4 arg AM request w/ I.
 * IN:		reply_endpoint	-> endpoint index to send to.
 *			handler			-> The index into the remote handler table.
 *			source			-> Source bytes to send.
 *			nbytes			-> size of source.
 *			a0..a3			-> The free args.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An AM is posted to the bundle.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::RequestI4(int reply_endpoint, handler_t handler, void *source, int nbytes, int a0, int a1, int a2, int a3)
{
	cConnection *connection;
	cPacket		*packet;
	Header		*header;

	connection = mMap[reply_endpoint].connection;
	if(!connection)
	{
		return AM_ERR_BAD_ARG;
	}

	// Get a free packet.
	packet = connection->GetFreePacket();
	if(!packet)
	{
		AM_Error("Unable to get free packet for send.", __LINE__, __FILE__, AM_ERR_RESOURCE);
		return AM_ERR_RESOURCE;
	}

	// Set up the packet.
	packet->SetData(source, nbytes);

	// Set up the header.
	header = packet->GetHeader();
	header->opcode = _AM_REQUEST;
	header->handlerType = AM_REQUEST_I4;
	header->tag = mMap[reply_endpoint].tag;
	header->args[0] = a0;
	header->args[1] = a1;
	header->args[2] = a2;
	header->args[3] = a3;
	header->destEndpoint = mMap[reply_endpoint].number;
	header->srcEndpoint  = mNumber;
	header->handlerNum = handler;
	header->nSize = nbytes;
	return connection->SendPacket(packet);
}

/*
 * cEndpoint::RequestXfer4()
 *
 * Purpose:	Does a 4 arg AM request w/ xfer.
 * IN:		reply_endpoint	-> endpoint index to send to.
 *			dest_offset		-> offset in remote endpoints VM segment.
 *			handler			-> The index into the remote handler table.
 *			source			-> Source bytes to send.
 *			nbytes			-> size of source.
 *			a0..a3			-> The free args.
 * OUT:		-
 * Cond:	-
 * PostCnd:	An AM is posted to the bundle.
 * Return:	AM_OK or AM_ERR_* otherwise.
 */
int cEndpoint::RequestXfer4(int reply_endpoint, int dest_offset, handler_t handler, void *source, int nbytes, int a0, int a1, int a2, int a3)
{
	cConnection *connection;
	cPacket		*packet;
	Header		*header;

	int			 bytesRemaining = nbytes;
	int			 offset		    = 0;

	connection = mMap[reply_endpoint].connection;
	if(!connection)
	{
		return AM_ERR_BAD_ARG;
	}

	// Fragment into AM_MAX_MEDIUM size chunks.
	while(bytesRemaining > 0)
	{

		// Get a free packet.
		packet = connection->GetFreePacket();
		if(!packet)
		{
			AM_Error("Unable to get free packet for send.", __LINE__, __FILE__, AM_ERR_RESOURCE);
			return AM_ERR_RESOURCE;
		}

		// Get packet header.
		header = packet->GetHeader();

		// Set up part of header.
		if(bytesRemaining > AM_MAX_MEDIUM)
		{
			header->opcode = _AM_XFER_INTERMEDIATE;
			packet->SetData((char *)source+offset, AM_MAX_MEDIUM);
			header->nSize       = AM_MAX_MEDIUM;
			header->destOffset = dest_offset+offset;
		}
		else
		{
			header->opcode = _AM_REQUEST;
			packet->SetData((char *)source+offset, bytesRemaining);
			header->handlerType = AM_REQUEST_XFER4;
			header->nSize		= nbytes;
			header->destOffset	= dest_offset;
		}

		// Set up rest of header..
		header->tag = mMap[reply_endpoint].tag;
		header->args[0] = a0;
		header->args[1] = a1;
		header->args[2] = a2;
		header->args[3] = a3;
		header->destEndpoint = mMap[reply_endpoint].number;
		header->srcEndpoint  = mNumber;
		header->handlerNum = handler;
		if(connection->SendPacket(packet) != AM_OK)
		{
			printf("Error, unable to send XFER!\n");
		}

		offset+=header->nSize;
		bytesRemaining -= header->nSize;
	}
	return AM_OK;
}


/*
 * cEndpoint::ProcessXferIntermediate()
 *
 * Purpose:	Processes an intermediate transfer.
 * IN:		packet	-> The incoming packet.
 * OUT:		-
 * Cond:	-
 * PostCnd:	
 * Return:	AM_OK if success.
 */
int	cEndpoint::ProcessXferIntermediate(cPacket *packet)
{
	Header *header = packet->GetHeader();
	memcpy((char *)mSegment+header->destOffset, packet->GetData(), AM_MAX_MEDIUM);
	return AM_OK;
}

/*
 * cEndpoint::ProcessRequest()
 *
 * Purpose:	Processes a Request
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	
 * Return:	
 */
int	cEndpoint::ProcessRequest(cPacket *packet)
{
	Header	   *header;
	Token		token;
	void	   *function;
	void	   *payload;
	int			size;
	int			offset;

	header = packet->GetHeader();

//	printf("-----------------AM Details-----------------\n");
//	printf("Opcode:			%d (_AM_REQUEST)\n", header->opcode);
//	printf("Return Address: %s:%s:%s:%d\n", header->token.source.node, 
//										   header->token.source.unique,
//										   header->token.source.bundle,
//										   header->token.source.endpoint);
//	printf("Dest Endpoint:  %d\n", header->destEndpoint);
//	printf("Dest Handler:   %d\n", header->handlerNum);
//	printf("Handler Type:   %d\n", header->handlerType);
//	printf("Payload Size:	%d\n", header->nSize);
//	printf("--------------------------------------------\n");

	// Make sure the function is OK.
	function = GetFunction(header->handlerNum);
	if(function == abort)
	{
		header->status = EBADHANDLER;
		return AM_ERR_RESOURCE;			// RTS: EABADHANDLER
	}

	// Do tag match.
	if(mTag != AM_ALL)
	{
		if((mTag == AM_NONE) || (mTag != header->tag))
		{
			printf("Returning packet to sender...\n");
			header->status = EBADTAG;
			return AM_ERR_RESOURCE;		// RTS:	EBADTAG
		}
	}

	// Fill in the endpoint address
	token.endp		 = this;
	token.connection = header->hidden.connPtr;
	token.tag		 = header->tag;
	token.source	 = header->srcEndpoint;

	switch(header->handlerType)
	{
		case AM_REQUEST_4:
			// Call the requested handler.
			((Handler4)function)(&token, header->args[0], header->args[1], header->args[2], header->args[3]);	
			break;

		case AM_REQUEST_I4:
			payload = packet->GetData();

			// Call the requested handler.
			((Handler4I)function)(&token, payload, header->nSize, header->args[0], header->args[1], header->args[2], header->args[3]);
			break;

		case AM_REQUEST_XFER4:
			payload		= packet->GetData();
			size		= ((header->nSize % AM_MAX_MEDIUM) == 0) ? AM_MAX_MEDIUM : header->nSize % AM_MAX_MEDIUM;
			offset		= header->destOffset + header->nSize - size;
			memcpy((char *)mSegment+offset, payload, size);

			// Call the registered handler.
			((Handler4I)function)(&token, (char *)mSegment+header->destOffset, header->nSize, header->args[0], header->args[1], header->args[2], header->args[3]);
			break;

		default:
				AM_Error("Unknown request handler type.", __LINE__, __FILE__);
	} /* switch */

	return AM_OK;
}

/*
 * cEndpoint::ProcessReply()
 *
 * Purpose:	Processes a Reply
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	
 * Return:	
 */
int	cEndpoint::ProcessReply(cPacket *packet)
{
	Header	   *header;
	Token		token;
	void	   *function;
	void	   *payload;
	int			size;
	int			offset;

	header = packet->GetHeader();

//	printf("-----------------AM Details-----------------\n");
//	printf("Opcode:			%d\n", header->opcode);
//	printf("Return Address: %s:%s:%s:%d\n", header->token.source.node, 
//										   header->token.source.unique,
//										   header->token.source.bundle,
//										   header->token.source.endpoint);
//	printf("Dest Endpoint:  %d\n", header->destEndpoint);
//	printf("Dest Handler:   %d\n", header->handlerNum);
//	printf("Handler Type:   %d\n", header->handlerType);
//	printf("Payload Size:	%d\n", header->nSize);
//	printf("--------------------------------------------\n");

	// Make sure the function is OK
	function = GetFunction(header->handlerNum);
	if(function == abort)
	{
		header->status = EBADHANDLER;	
		return AM_ERR_RESOURCE;			// RTS: EABADHANDLER
	}

	// Do tag match.
	if(mTag != AM_ALL)
	{
		if((mTag == AM_NONE) || (mTag != header->tag))
		{
			header->status = EBADTAG;
			return AM_ERR_RESOURCE;		// RTS:	EBADTAG
		}
	}

	// Fill in the endpoint address
	token.endp		 = this;
	token.connection = header->hidden.connPtr;
	token.tag		 = header->tag;
	token.source	 = header->srcEndpoint;

	switch(header->handlerType)
	{
		case AM_REPLY_4:
			// Call the reply handler.
			((Handler4)function)(&token, header->args[0], header->args[1], header->args[2], header->args[3]);	
			break;

		case AM_REPLY_I4:
			// Call the reply handler.
			((Handler4I)function)(&token, packet->GetData(), header->nSize, header->args[0], header->args[1], header->args[2], header->args[3]);
			break;

		case AM_REPLY_XFER4:
			payload		= packet->GetData();
			size		= ((header->nSize % AM_MAX_MEDIUM) == 0) ? AM_MAX_MEDIUM : header->nSize % AM_MAX_MEDIUM;
			offset		= header->destOffset + header->nSize - size;
			memcpy((char *)mSegment+offset, payload, size);

			// Call the registered handler.
			((Handler4I)function)(&token, (char *)mSegment+header->destOffset, header->nSize, header->args[0], header->args[1], header->args[2], header->args[3]);
			break;

		default:
			AM_Error("Unknown reply handler type.", __LINE__, __FILE__);
	}
	return AM_OK;
}


/*
 * cEndpoint::ProcessRTS()
 *
 * Purpose:	Processes a return to sender packet.
 * IN:		packet		-> The packet to be returned to the sender.
 * OUT:		-
 * Cond:	-
 * PostCnd:	
 * Return:	AM_OK else AM_ERR_*
 */
int	cEndpoint::ProcessRTS(cPacket *packet)
{
	Header	   *header;
	Token		token;
	void	   *function;
	_argblock_t	argblock;
	int			i;

	header = packet->GetHeader();


//	printf("-----------------AM <Return-to-Sender> Details-----------------\n");
//	printf("Opcode:			%d\n", header->opcode);
//	printf("Return Address: %s:%s:%s:%d\n", header->token.source.node, 
//										   header->token.source.unique,
//										   header->token.source.bundle,
//										   header->token.source.endpoint);
//	printf("Dest Endpoint:  %d\n", header->destEndpoint);
//	printf("Dest Handler:   %d\n", header->handlerNum);
//	printf("Handler Type:   %d\n", header->handlerType);
//	printf("Payload Size:	%d\n", header->nSize);
//	printf("--------------------------------------------\n");

	function = GetFunction(0);
	if(function == abort)
	{
		return AM_ERR_RESOURCE;
	}

	// Fill in the endpoint address
	token.endp		 = this;
	token.connection = header->hidden.connPtr;
	token.tag		 = header->tag;
	token.source	 = header->srcEndpoint;

	argblock.handler		= header->handlerNum;
	argblock.reply_endpoint = header->destEndpoint;
	argblock.token			= &token;

	switch(header->handlerType)
	{
		case AM_REPLY_4:
		case AM_REQUEST_4:
			for(i = 0; i < 4; i++) 
			{
				argblock.args[i] = header->args[i];
			}
			argblock.dest_offset = 0;
			argblock.src_offset  = 0;
			argblock.nbytes		 = 0;
			argblock.source_addr = 0;
			break;

		case AM_REQUEST_I4:
		case AM_REPLY_I4:
			for(i = 0; i < 4; i++) 
			{
				argblock.args[i] = header->args[i];
			}
			argblock.dest_offset = 0;
			argblock.src_offset  = 0;
			argblock.nbytes		 = header->nSize;
			argblock.source_addr = packet->GetData();
			break;

		default:
			AM_Error("Unknown reply handler type.", __LINE__, __FILE__);
	}

	((Undeliverable)function)((unsigned int)header->status, (unsigned int)header->handlerType, &argblock);

	return AM_OK;
}

/*
 * cEndpoint::GetNumber()
 *
 * Purpose:	Gets the unique number that identifies this ep (within a bundle).
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	The endpoint number.
 */
int	cEndpoint::GetNumber()
{
	return mNumber;
}

/*
 * cEndpoint::GetBundle()
 *
 * Purpose:	Returns a pointer to the bundle this belongs to.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	A pointer to the cBundle object.
 */
cBundle* cEndpoint::GetBundle()
{
	return mBundle;
}

//Std abort function.
void abort()
{
	printf("Abort!\n");
}

/*
 * cEndpoint::GetAddress()
 *
 * Purpose:	Get a pointer to this endpoint's address.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	A pointer to the endpoint's unique address.
 */
en_t cEndpoint::GetAddress()
{
	return &mAddress;
}

/*
 * cEndpoint::SetAddress()
 *
 * Purpose:	Sets this endpoints unique address.
 * IN:		addr	-> A pointer to the address that should be copied.
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	-
 */
void cEndpoint::SetAddress(_en_t *addr)
{
	memcpy(&mAddress, addr, sizeof(_en_t));	
}

/*
 * cEndpoint::GetTranslationName()
 *
 * Purpose:	Fills gan with a pointer to the remote endpoint for the given index.
 * IN:		i		-> table entry num
 * OUT:		gan		-> Pointer to the remote endpoint's name.
 * Cond:	-
 * PostCnd:	-
 * Return:	AM_OK or AM_ERR_*
 */
int cEndpoint::GetTranslationName(int i, en_t *gan)
{
	if(mMap[i].connection == NULL)
	{
		return AM_ERR_BAD_ARG;
	}
	else
	{
		*gan = mMap[i].connection->GetAMAddress();
		return AM_OK;
	}
}