/*
 * cPacket.cpp
 * 
 * Class implementatino for AM packet class. 
 *
 * Authors:  Vicky & Ted
 * Last Mod: 4 March 1999 by ted
 */
#include "cPacket.h"

/*
 * cPacket::cPacket()
 *
 * Purpose:	Create and initialize a new packet.
 * IN:		dataSize	-> The size of the data segment.
 *			nic			-> Pointer to the cNic object.
 *			ptag		-> Pointer to the cPtag object.
 * OUT:		-
 * Cond:	The AM Layer is initialized.
 * PostCnd:	
 * Return:	-
 */
cPacket::cPacket(int dataSize, cNic *nic, cPtag *ptag)
{

	VIP_MEM_ATTRIBUTES	memAttrib;
	VIP_RETURN			rc;
	void			   *mDataPtr = NULL;

	mNic			= nic->GetNicHandle();
	mHeader			= NULL;
	mDescAllocPtr	= NULL;
	mDataSize = dataSize;
	
	// Set up memory attributes.
	memAttrib.Ptag			  = ptag->GetPtag();
	memAttrib.EnableRdmaWrite = VIP_FALSE;
	memAttrib.EnableRdmaRead  = VIP_FALSE;

	// Alloc data.
	mDataPtr = calloc(1, sizeof(Header)+dataSize);
	if(!mDataPtr)
	{
		throw AMException("Couldn't allocate data in packet constructor.");
	}
	mHeader			= (Header *)mDataPtr;
	mData			= (char *)mDataPtr+sizeof(Header);

	// Alloc descriptor.
	mDescAllocPtr = calloc(1, sizeof(VIP_DESCRIPTOR)+VIP_DESCRIPTOR_ALIGNMENT);
	if(!mDescAllocPtr)
	{
		throw AMException("Couldn't allocate memory for Desc.");
	}
	mDescPtr = (VIP_DESCRIPTOR *)
				((int)((char *)mDescAllocPtr + VIP_DESCRIPTOR_ALIGNMENT - 1)&(0-VIP_DESCRIPTOR_ALIGNMENT));

	// Register the descriptor.
	rc = VipRegisterMem(mNic, mDescPtr, sizeof(VIP_DESCRIPTOR), &memAttrib, &mDescMemHandle);
	if (rc != VIP_SUCCESS)
	{
		throw AMException("Couldn't register memory for descriptor.");
	}
	
	// register the memory for data/header.
	rc = VipRegisterMem(mNic, (void *)mHeader, sizeof(Header), &memAttrib, &mHeaderMemHandle);
	if (rc != VIP_SUCCESS)
	{
		throw AMException("Couldn't register memory in packet constructor.");
	}
	rc = VipRegisterMem(mNic, mData, dataSize, &memAttrib, &mDataMemHandle);
	if (rc != VIP_SUCCESS)
	{
		throw AMException("Couldn't register memory in packet constructor.");
	}

	// Set up descriptor
	mDescPtr->CS.Control = VIP_CONTROL_OP_SENDRECV;
	GetHeader()->hidden.packetPtr = this;

	// Set the segment count.
	mDescPtr->CS.SegCount = 2;

	// Set segment[0]
	mDescPtr->DS[0].Local.Data.Address	= mHeader;
	mDescPtr->DS[0].Local.Handle		= mHeaderMemHandle;
	mDescPtr->DS[0].Local.Length		= sizeof(Header)-sizeof(HiddenData);

	// Set segment[1]
	mDescPtr->DS[1].Local.Data.Address	= mData;
	mDescPtr->DS[1].Local.Handle		= mDataMemHandle;
	mDescPtr->DS[1].Local.Length		= mDataSize;

	mDescPtr->CS.Length = sizeof(Header)-sizeof(HiddenData)+mDataSize;
}

/*
 * cPacket::~cPacket()
 *
 * Purpose:	Destroys the packet.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	-
 */
cPacket::~cPacket()
{
	VIP_RETURN rc;

	// Free the header and data segments.
	if(mHeader)
	{
		rc = VipDeregisterMem(mNic, mHeader, mHeaderMemHandle);
		if(rc != VIP_SUCCESS)
		{
			AM_Error("Unable to deregister header segment.", __LINE__, __FILE__);
		}
		rc = VipDeregisterMem(mNic, mData, mDataMemHandle);
		if(rc != VIP_SUCCESS)
		{
			AM_Error("Unable to deregister data segment.", __LINE__, __FILE__);
		}
		free((void *)mHeader);
	}

	// Free the descriptor.
	if(mDescAllocPtr)
	{
		rc = VipDeregisterMem(mNic, mDescPtr, mDescMemHandle);
		if(rc != VIP_SUCCESS)
		{
			AM_Error("Couldn't deregister memory in for desc.", __LINE__, __FILE__);
		}
		free(mDescAllocPtr);
	}
}

/*
 * cPacket::Reset()
 *
 * Purpose:	Resets the packet to normal.
 * IN:		-
 * OUT:		-
 * Cond:	-
 * PostCnd:	-
 * Return:	-
 */
void cPacket::Reset()
{
	// Set the segment count.
	mDescPtr->CS.SegCount = 2;

	// Set segment[0]
	mDescPtr->DS[0].Local.Length       = sizeof(Header)-sizeof(HiddenData);

	// Set segment[1]
	mDescPtr->DS[1].Local.Length       = mDataSize;

	// Set overall length.
	mDescPtr->CS.Length = sizeof(Header)-sizeof(HiddenData)+mDataSize;
}

inline int  cPacket::GetLength()
{
	return ((mDescPtr->CS.Length)-(sizeof(Header)-sizeof(HiddenData)));
}

void cPacket::SetData(void *buf, int length)
{
	memcpy(mData, buf, length);
	mDescPtr->DS[1].Local.Length = length;
	SetLength(length);
}

