/*
 * cNic.cpp
 * 
 * Class implementation for VI NIC (Network Interface Connection) object.
 *
 * Authors:  Vicky & Ted
 * Last Mod: 4 March 1999 by ted
 */
#include "stdvint.h"
#include "cNic.h"

#include "cVI.h"
#include "cDesc.h"
#include "cDSeg.h"
#include "cPtag.h"
#include "cVIAddr.h"
#include "cCQ.h"

//cNic class

//Constructor(s)
cNic::cNic(const char *DeviceName)
{
	VIP_RETURN	rc;

	mOpen = VIP_FALSE;

	// Open the Nic
	rc = VipOpenNic(DeviceName, &mNic);
	if (rc != VIP_SUCCESS)
	{
		throw VIException("Unable to open Nic.");
	}

	// Initialize Name Service using default provider
	rc = VipNSInit(mNic, (VIP_PVOID)0); 
	if(rc != VIP_SUCCESS)
	{
		throw VIException("Unable to initialize the name service.");
	}

	// Query the Nic Attributes.
	rc = VipQueryNic(mNic, &mAttrib);
	if(rc != VIP_SUCCESS)
	{
		throw VIException("Unable to query nic.");
	}

	mOpen = VIP_TRUE;
}

//Destructor
cNic::~cNic()
{
	VIP_RETURN	rc;	

	// Shut down name service.
	rc = VipNSShutdown(mNic);
	if (rc != VIP_SUCCESS)
	{
		cVIError::PrintError("Unable to shut down the name service.", __LINE__, __FILE__, rc);
	}

	// Close the Nic.
	rc = VipCloseNic(mNic);
	if (rc != VIP_SUCCESS)
	{
		cVIError::PrintError("Unable to close the nic.", __LINE__, __FILE__, rc);
	}

}


//Accessor Functions
const VIP_NIC_ATTRIBUTES*	cNic::GetAttrib()
{
	return &mAttrib;
}
VIP_BOOLEAN					cNic::IsOpen()
{
	return mOpen;
}
	

//Set error callback.
VIP_RETURN	cNic::ErrorCallback(VIP_PVOID Context,
										  void(*Handler) (
													VIP_PVOID			 Context,
													VIP_ERROR_DESCRIPTOR *ErrorDesc
													)
										  )
{
	VIP_RETURN rc;

	rc = VipErrorCallback(mNic, Context, Handler);
	if (rc != VIP_SUCCESS)
	{
		cVIError::PrintError("Unable to set error callback.", __LINE__, __FILE__, rc);
	}
	return rc;
}



//Object Creation Functions
cVI*	cNic::CreateVI(VIP_VI_ATTRIBUTES *attrib)
{
	cVI		*out = NULL;

	try 
	{
		out = new cVI(mNic, attrib);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}
	return out;
}

cVI*	cNic::CreateVI(VIP_VI_ATTRIBUTES *attrib, cCQ *SendCQ, cCQ *RecvCQ)
{
	cVI		*out = NULL;

	try 
	{
		out = new cVI(mNic, attrib, SendCQ, RecvCQ);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}
	return out;
}

cCQ*	cNic::CreateCQ(VIP_ULONG size)
{
	cCQ		*out = NULL;

	try 
	{
		out = new cCQ(size, mNic);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}
	return out;
}


cDesc*	cNic::CreateDesc(int numSegments, cPtag *ptag, VIP_BOOLEAN RdmaWrite)
{
	cDesc	*out = NULL;
	VIP_MEM_ATTRIBUTES	memAttrib;

	memAttrib.Ptag			  = ptag->GetPtag();
	memAttrib.EnableRdmaWrite = RdmaWrite;
	memAttrib.EnableRdmaRead  = VIP_FALSE;

	try
	{
		out = new cDesc(numSegments, &memAttrib, mNic);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}

	return out;
}

cDSeg*	cNic::CreateDSeg(VIP_ULONG length, cPtag *ptag, VIP_BOOLEAN RdmaWrite, VIP_BOOLEAN RdmaRead)
{
	cDSeg					*out = NULL;
	VIP_MEM_ATTRIBUTES		memAttrib;

	memAttrib.Ptag			  = ptag->GetPtag();
	memAttrib.EnableRdmaWrite = RdmaWrite;
	memAttrib.EnableRdmaRead  = RdmaRead;

	try
	{
		out = new cDSeg(length, &memAttrib, mNic);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}
	return out;
}

cDSeg*	cNic::CreateDSeg(VIP_ULONG length, void *buffer, cPtag *ptag, VIP_BOOLEAN RdmaWrite, VIP_BOOLEAN RdmaRead)
{
	cDSeg					*out = NULL;
	VIP_MEM_ATTRIBUTES		memAttrib;

	memAttrib.Ptag			  = ptag->GetPtag();
	memAttrib.EnableRdmaWrite = RdmaWrite;
	memAttrib.EnableRdmaRead  = RdmaRead;

	try
	{
		out = new cDSeg(length, buffer, &memAttrib, mNic);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}
	return out;
}

cPtag*   cNic::CreatePtag()
{
	cPtag	*out = NULL;
	try
	{
		out = new cPtag(mNic);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}

	return out;
}


cVIAddr*	cNic::CreateVIAddr()
{
	cVIAddr	*out = NULL;

	try
	{
		out = new cVIAddr(mNic, mAttrib.NicAddressLen, mAttrib.MaxDiscriminatorLen);
	}
	catch(VIException &e)
	{
		cVIError::PrintError(e.ToString(), __LINE__, __FILE__);
		delete out;
		out = NULL;
	}

	return out;
}
