/*
 * cVIAddr.cpp
 * 
 * Class implementation for VI addresses. 
 *
 * Authors:  Vicky & Ted
 * Last Mod: 27 Feb 1999
 */
#include "cVIAddr.h"

#define  MAX_LOCAL_NAME_SIZE	255

	
//Constructor(s)
cVIAddr::cVIAddr(VIP_NIC_HANDLE nicHandle, int nicAddressLen, int nicMaxDiscLen)
{
	mIsValid		= VIP_FALSE;
	mNic			= nicHandle;
	mMaxDiscLen	    = nicMaxDiscLen;
	mAddr = NULL;

	mAddr = (VIP_NET_ADDRESS *) calloc(1, 
									   sizeof(VIP_NET_ADDRESS)+
		                               (nicAddressLen + mMaxDiscLen));
	if(!mAddr)
	{
		throw VIException("Unable to alloc space for address in cVIAddr::cVIAddr().");
	}
	mAddr->HostAddressLen   = nicAddressLen;
	mAddr->DiscriminatorLen = 0;
}	

//Copy constructor
cVIAddr::cVIAddr(const cVIAddr &original)
{
	mIsValid	= original.mIsValid;
	mNic		= original.mNic;
	mMaxDiscLen = original.mMaxDiscLen;

	mAddr = (VIP_NET_ADDRESS *) calloc(1, sizeof(original.mAddr));
	if(!mAddr)
	{
		throw VIException("Unable to alloc space for address in cVIAddr copy constructor.");
	}
	memcpy(mAddr, original.mAddr, sizeof(original.mAddr));
}

//Destructor
cVIAddr::~cVIAddr()
{
	if(mAddr)
	{
		free(mAddr);	
	}
}

//Create a copy of this cVIAddr object.
cVIAddr* cVIAddr::Copy()
{
	cVIAddr	*ret;

	ret = new cVIAddr(*this);
	if (!ret)
	{
		return NULL;
	} 
	else
	{
		return ret;
	}
}


//-----------Useful functions-------------//

/*
 * GetLocalAddress
 * 
 * Attempts to fill this address object based on the given name string.
 *
 * Ret  : VIP_TRUE if success, VIP_FALSE otherwise.
 * Side : isValid is set if success, false otherwise.
 */
VIP_BOOLEAN cVIAddr::GetHostByName(VIP_CHAR *name)
{
	VIP_RETURN	rc;

	rc = VipNSGetHostByName(mNic, name, mAddr, 0);
	if(rc != VIP_SUCCESS)
	{
		mIsValid = VIP_FALSE;
	}
	else
	{
		mIsValid = VIP_TRUE;
	}
	return mIsValid;
	
} /* end GetHostByName */


/*
 * GetLocalAddress
 * 
 * Attempts to fill this address object w/ the local address.
 *
 * Ret  : VIP_TRUE if success, VIP_FALSE otherwise.
 * Side : isValid is set if success, false otherwise.
 */
VIP_BOOLEAN cVIAddr::GetLocalAddress()
{
	char localName[MAX_LOCAL_NAME_SIZE];
	unsigned long nsize	= MAX_LOCAL_NAME_SIZE;

	BOOL bName = GetComputerName(localName, &nsize);
	if(bName != TRUE)
	{
		mIsValid = VIP_FALSE;
	}
	else
	{
		mIsValid = GetHostByName(localName);
	}
	return mIsValid;
} /* end GetLocalAddress */


/*
 * GetLocalNodeName
 * 
 * Attempts to fill the provided string buffer w/ the local node name.
 * In	 : name	  -> a buffer in which to store the node name.
 *		   length -> a pointer to an integer containing the length of the buffer.\
 * Out	 : length -> gets updated with the length of the name.
 * Ret   : VIP_TRUE if success, VIP_FALSE otherwise.
 * Side  : isValid is set if success, false otherwise.
 */
VIP_BOOLEAN cVIAddr::GetLocalNodeName(char *name, unsigned long *length)
{
	BOOL bName = GetComputerName(name, length);
	if(bName != TRUE)
	{
		return VIP_FALSE;
	}
	else
	{
		return VIP_TRUE;
	}
}

/*
 * GetHostByAddr
 * 
 * Attempts to fill the given string with the name of the node.
 *
 * Ret  : VIP_TRUE if success, VIP_FALSE otherwise.
 * Side : outName is a null-terminated string w/ the node name.
 *        outNameLen is the size of outName (not including NULL term)
 *	      isValid is set if success, false otherwise.
 */
VIP_BOOLEAN cVIAddr::GetHostByAddr(cVIAddr *addr, VIP_CHAR *outName, VIP_ULONG *outNameLen)
{
	VIP_RETURN			 rc;

	// Make sure given a "valid" address.
	if(!(addr->isValid()))
	{
		return VIP_FALSE;
	}

	rc = VipNSGetHostByAddr(addr->mNic, addr->GetAddress(), outName, outNameLen);
	if (rc != VIP_SUCCESS)
	{
		return VIP_FALSE;
	}

	// Success
	return VIP_TRUE;
	
}


VIP_BOOLEAN cVIAddr::GetHostByAddr(VIP_CHAR *outName, VIP_ULONG *outNameLen)
{
	return GetHostByAddr(this, outName, outNameLen);	
}

//Accessor Functions
VIP_NET_ADDRESS *cVIAddr::GetAddress()
{
	return mAddr;
}

VIP_BOOLEAN	 cVIAddr::isValid()
{
	return mIsValid;	
}

VIP_BOOLEAN	cVIAddr::setValid(VIP_BOOLEAN valid)
{
	mIsValid = valid;
	return mIsValid;
}


/*
 * SetDiscriminator
 * 
 * Fills in the discriminator w/ length bytes from the given string.
 *
 * Ret  : VIP_TRUE if success, VIP_FALSE otherwise.
 * Side : addr->DiscriminatorLen is set.
 */
VIP_BOOLEAN cVIAddr::SetDiscriminator(const char *disc, int length) 
{

	// Make sure setting legal discriminator length.
	if(length > mMaxDiscLen)
	{
		cVIError::PrintError("Attempt to set discriminator larger than max disc. length.", __LINE__, __FILE__);
		return VIP_FALSE;
	}
	
	mAddr->DiscriminatorLen = length;
	memcpy(mAddr->HostAddress+mAddr->HostAddressLen, disc, length);
	return VIP_TRUE;
}


/*
 * GetDiscriminator
 * 
 * Fills in the given string with the discriminator.  Fills in the
 * given length integer w/ the length of the discriminator.
 *
 * Param : disc <- A buffer.   length <- the size of the buffer.
 * Out	 : disc <- the discriminator.  length <- the size of disc.
 * Ret   : VIP_TRUE if success, VIP_FALSE otherwise.
 * Side  : None.
 */
VIP_BOOLEAN cVIAddr::GetDiscriminator(char *disc, int *length)
{
	
	// Make sure the buffer size given isn't too small.
	if(*length < mAddr->DiscriminatorLen)
	{
		cVIError::PrintError("cVIAddr::GetDiscriminator -> Given buffer is too small.", __LINE__, __FILE__);
		return VIP_FALSE;
	}

	*length = mAddr->DiscriminatorLen;
	memcpy(disc, mAddr->HostAddress+mAddr->HostAddressLen, *length);
	return VIP_TRUE;
}
